"""Part of the module for electrodynamic simulations

"""

# flake8: noqa

from ase.units import Hartree, Bohr
from gpaw.fd_operators import Gradient
import numpy as np

# in atomic units, 1/(4*pi*e_0) = 1
_eps0_au = 1.0 / (4.0 * np.pi)

# Base class for the classical polarizable material:
#    -holds various functions at each point in the calculation grid:
#        1) the dielectric function (permittivity)
#        2) electric field
#        3) classical polarization charge density
#    -contains routines for calculating them from each other and/or external potential

class PolarizableMaterial():
    def __init__(self, *, sign=-1.0):
        self.gd          = None
        self.initialized = False
        self.sign        = sign
        self.messages    = []
        self.components = []

    def add_component(self, component):
        self.components.append(component)

    def permittivity_value(self, omega=0.0):
        return self.eps_infty + _eps0_au * np.sum(self.beta / (self.bar_omega**2.0 - 1J * self.alpha * omega - omega**2.0), axis=0)

    def get_static_permittivity(self):
        return self.gd.collect(self.permittivity_value(0.0))

    def initialize(self, gd):
        if self.initialized: # double initialization leads to problems
            return
        self.initialized = True
        self.messages.append("Polarizable Material:")

        try:
            self.Nj = max(component.permittivity.Nj for component in self.components)
        except:
            self.Nj = 0

        self.gd = gd

        # 3-dimensional scalar array: rho, eps_infty
        self.charge_density = self.gd.zeros()
        self.eps_infty = np.ones(self.gd.empty().shape) * _eps0_au

        # 3-dimensional vector arrays:
        #        electric field, total polarization density
        dims = [3] + list(self.gd.empty().shape)
        self.electric_field = np.zeros(dims)
        self.polarization_total = np.zeros(dims)

        # 4-dimensional vector arrays:
        #        currents, polarizations
        dims = [3, self.Nj] + list(self.gd.empty().shape)
        self.currents      = np.zeros(dims)
        self.polarizations = np.zeros(dims)

        # 4-dimensional scalar arrays:
        #        oscillator parameters alpha, beta, bar_omega, eps_infty
        dims = [self.Nj] + list(self.gd.empty().shape)
        self.alpha     = np.zeros(dims)
        self.beta      = np.zeros(dims)
        self.bar_omega = np.ones(dims)

        # Set the permittivity for each grid point
        for component in self.components:
            self.apply_mask(mask = component.get_mask(self.gd),
                           permittivity = component.permittivity)

    # Restart by regenerating the structures
    # TODO: is everything always fully reproduced? Should there be a test for it (e.g. checksum[mask])
    def read(self, reader):
        for component in reader.classmat.components:
            name = component['name']
            arguments = component['arguments']
            eps_infty = component['eps_infty']
            eps = component['eps']  # Data as array
            eval('self.add_component(%s(permittivity = Permittivity(data=eps), %s))' % (name, arguments))

    # Save information on the structures for restarting
    def write(self, writer):
        writer.write(components=[
                         {'name': component.name,
                          'arguments': component.arguments,
                          'eps_infty': component.permittivity.eps_infty,
                          'eps': component.permittivity.data_eVA()}
                         for component in self.components])

    # Here the 3D-arrays are filled with material-specific information
    def apply_mask(self, mask, permittivity):
        for j in range(permittivity.Nj):
            self.bar_omega[j] = np.logical_not(mask) * self.bar_omega[j] + mask * permittivity.oscillators[j].bar_omega
            self.alpha    [j] = np.logical_not(mask) * self.alpha[j]    + mask * permittivity.oscillators[j].alpha
            self.beta     [j] = np.logical_not(mask) * self.beta[j]     + mask * permittivity.oscillators[j].beta

        # Add dummy oscillators if needed
        for j in range(permittivity.Nj, self.Nj):
            self.bar_omega  [j] = np.logical_not(mask) * self.bar_omega[j] + mask * 1.0
            self.alpha      [j] = np.logical_not(mask) * self.alpha[j]    + mask * 0.0
            self.beta       [j] = np.logical_not(mask) * self.beta[j]     + mask * 0.0

        # Print the permittivity information
        self.messages.append("  Permittivity data:")
        self.messages.append("    bar_omega         alpha          beta")
        self.messages.append("  ----------------------------------------")
        for j in range(permittivity.Nj):
            self.messages.append("%12.6f  %12.6f  %12.6f" % (permittivity.oscillators[j].bar_omega,
                                                             permittivity.oscillators[j].alpha,
                                                             permittivity.oscillators[j].beta))
        self.messages.append("  ----------------------------------------")
        self.messages.append("...done initializing Polarizable Material")
        masksum  = self.gd.comm.sum(int(np.sum(mask)))
        masksize = self.gd.comm.sum(int(np.size(mask)))
        self.messages.append("Fill ratio: %f percent" % (100.0 * float(masksum)/float(masksize)))


    # E(r) = -Grad V(r)
    # NB: Here -V(r) is used (if/when sign=-1), because in GPAW the
    #     electron unit charge is +1 so that the calculated V(r) is
    #     positive around negative charge. In order to get the correct
    #     direction of the electric field, the sign must be changed.
    def solve_electric_field(self, phi):
        for v in range(3):
            Gradient(self.gd, v, n=3).apply(-1.0 * self.sign * phi, self.electric_field[v])

    # n(r) = -Div P(r)
    def solve_rho(self):
        self.charge_density *= 0.0
        dmy         = self.gd.empty()
        for v in range(3):
            Gradient(self.gd, v, n=3).apply(self.polarization_total[v], dmy)
            self.charge_density -= dmy

    # P(r, omega) = [eps(r, omega) - eps0] E(r, omega)
    # P0(r) = [eps_inf(r) - eps0] E0(r) + sum_j P0_j(r) // Gao2012, Eq. 10
    def solve_polarizations(self):
        for v in range(3):
            self.polarizations[v] = _eps0_au * self.beta / (self.bar_omega**2.0) * self.electric_field[v]
        self.polarization_total = np.sum(self.polarizations, axis=1) + (self.eps_infty - _eps0_au ) * self.electric_field

    def propagate_polarizations(self, timestep):
        for v in range(3):
            self.polarizations[v] = self.polarizations[v] + timestep * self.currents[v]
        self.polarization_total = np.sum(self.polarizations, axis=1)

    def propagate_currents(self, timestep):
        c1 = (1.0 - 0.5 * self.alpha*timestep)/(1.0 + 0.5 * self.alpha*timestep)
        c2 = - timestep / (1.0 + 0.5 * self.alpha*timestep) * (self.bar_omega**2.0)
        c3 = - timestep / (1.0 + 0.5 * self.alpha*timestep) * (-1.0) * _eps0_au * self.beta
        for v in range(3):
            self.currents[v] = c1 * self.currents[v] + c2 * self.polarizations[v] + c3 * self.electric_field[v]

    def kick_electric_field(self, timestep, kick):
        for v in range(3):
            self.electric_field[v] = self.electric_field[v] + kick[v] / timestep

# Box-shaped classical material
class PolarizableBox():
    def __init__(self, *, corner1, corner2, permittivity):
        # sanity check
        assert(len(corner1)==3)
        assert(len(corner2)==3)

        self.corner1      = np.array(corner1)/Bohr # from Angstroms to atomic units
        self.corner2      = np.array(corner2)/Bohr # from Angstroms to atomic units
        self.permittivity = permittivity

        self.name = 'PolarizableBox'
        self.arguments = 'corner1=[%f, %f, %f], corner2=[%f, %f, %f]' % (corner1[0], corner1[1], corner1[2],
                                                                         corner2[0], corner2[1], corner2[2])

    # Setup grid descriptor and the permittivity values inside the box
    def get_mask(self, gd):

        # 3D coordinates at each grid point
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))

        # inside or outside
        return np.logical_and(np.logical_and( # z
                np.logical_and(np.logical_and( # y
                 np.logical_and( #x
                                r_gv[:, :, :, 0] > self.corner1[0],
                                r_gv[:, :, :, 0] < self.corner2[0]),
                 r_gv[:, :, :, 1] > self.corner1[1]),
                               r_gv[:, :, :, 1] < self.corner2[1]),
                r_gv[:, :, :, 2] > self.corner1[2]),
                              r_gv[:, :, :, 2] < self.corner2[2])


# Shape from atom positions (surrounding region)
class PolarizableAtomisticRegion():
    def __init__(self, *, atoms=None, atom_positions=None, distance=0.0, permittivity=None):

        if atoms is not None:
            self.atom_positions = np.array(atoms.get_positions())
        else:
            self.atom_positions = np.array(atom_positions)

        # sanity check
        assert(len(self.atom_positions)>1)

        self.permittivity = permittivity
        self.name = 'PolarizableAtomisticRegion'

        # use the minimum interatomic distance
        if distance<1e-10:
            self.distance = np.sqrt(np.min([np.sort(np.sum((self.atom_positions-ap)**2, axis=1))[1] for ap in self.atom_positions]))/Bohr
        else:
            self.distance   = distance/Bohr # from Angstroms to atomic units

        dbohr = self.distance*Bohr
        self.arguments = 'distance = %20.12e, atom_positions=[' % dbohr
        for ap in self.atom_positions:
            self.arguments += '[%20.12e, %20.12e, %20.12e],' % (ap[0], ap[1], ap[2])
        self.arguments = self.arguments[:-1] + ']'

    # Setup grid descriptor and the permittivity values inside the box
    def get_mask(self, gd):

        # 3D coordinates at each grid point
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))

        # inside or outside
        mask_tbl = False * np.ones(r_gv.shape[:-1])
        for ap in self.atom_positions:
            mask_tbl = np.logical_or(mask_tbl, np.array( (r_gv[:, :, :, 0] - ap[0]/Bohr)**2.0 +
                                                         (r_gv[:, :, :, 1] - ap[1]/Bohr)**2.0 +
                                                         (r_gv[:, :, :, 2] - ap[2]/Bohr)**2.0 < self.distance**2))
        return mask_tbl

# Sphere-shaped classical material
class PolarizableSphere():
    def __init__(self, *, center, radius, permittivity):
        self.permittivity = permittivity
        self.center      = np.array(center)/Bohr # from Angstroms to atomic units
        self.radius      = radius/Bohr # from Angstroms to atomic units

        # sanity check
        assert(len(self.center)==3)

        self.name = 'PolarizableSphere'
        self.arguments = 'center=[%20.12e, %20.12e, %20.12e], radius=%20.12e' % (center[0], center[1], center[2], radius)

    def get_mask(self, gd):

        # 3D coordinates at each grid point
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))

        # inside or outside
        return  np.array( (r_gv[:, :, :, 0] - self.center[0])**2.0 +
                          (r_gv[:, :, :, 1] - self.center[1])**2.0 +
                          (r_gv[:, :, :, 2] - self.center[2])**2.0 < self.radius**2)

# Sphere-shaped classical material
class PolarizableEllipsoid():
    def __init__(self, *, center, radii, permittivity):
        # sanity check
        assert(len(center)==3)
        assert(len(radii)==3)

        self.center       = np.array(center)/Bohr # from Angstroms to atomic units
        self.radii        = np.array(radii)/Bohr   # from Angstroms to atomic units
        self.permittivity = permittivity

        self.name = 'PolarizableEllipsoid'
        self.arguments = 'center=[%20.12e, %20.12e, %20.12e], radii=[%20.12e, %20.12e, %20.12e]' % (center[0], center[1], center[2], radii[0], radii[1], radii[2])


    def get_mask(self, gd):

        # 3D coordinates at each grid point
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))

        # inside or outside
        return  np.array( (r_gv[:, :, :, 0] - self.center[0])**2.0/self.radii[0]**2.0 +
                          (r_gv[:, :, :, 1] - self.center[1])**2.0/self.radii[1]**2.0 +
                          (r_gv[:, :, :, 2] - self.center[2])**2.0/self.radii[2]**2.0 < 1.0)

 # Rod-shaped classical material
class PolarizableRod():
    def __init__(self, *, corners, radius, round_corners, permittivity):
        # sanity check
        assert(np.array(corners).shape[0]>1)  # at least two points
        assert(np.array(corners).shape[1]==3) # 3D

        self.name = 'PolarizableRod'
        self.arguments = 'radius = %20.12e, corners=[' % radius
        for c in corners:
            self.arguments += '[%20.12e, %20.12e, %20.12e],' % (c[0], c[1], c[2])
        self.arguments = self.arguments[:-1] + ']'
        self.arguments += f', round_corners={round_corners}'

        self.corners      = np.array(corners)/Bohr # from Angstroms to atomic units
        self.radius       = radius/Bohr  # from Angstroms to atomic units
        self.round_corners = round_corners
        self.permittivity = permittivity

    def get_mask(self, gd):

        # 3D coordinates at each grid point
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        ng = r_gv.shape[0:-1]
        ngv = r_gv.shape

        a = self.corners[0]

        mask = False * np.ones(ng)

        for p in self.corners[1:]:
            # http://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line:
            # d = |(a-p)-((a-p).n)n|   point p, line a+tn  (|n|=1)
            n = (p-a)/np.sqrt((p-a).dot((p-a)))
            v1 = np.array([a[w]-r_gv[:, :, :, w] for w in range(3)]).transpose((1, 2, 3, 0)) # a-p

            v2 = np.sum(np.array([v1[:, :, :, w]*n[w] for w in range(3)]), axis=0)  # (a-p).n

            v3 = np.array([v2*n[w] for w in range(3)]).transpose(1, 2, 3, 0)       # ((a-p).n)n

            d = np.zeros(ng)
            for ind, idx in np.ndenumerate(v3[:, :, :, 0]):
                d[ind] = np.array(v1[ind]-v3[ind]).dot(v1[ind]-v3[ind])

            # angle between (p-a) and (r-a):
            pa = p-a # (3)
            ra = np.array([r_gv[:, :, :, w]-a[w] for w in range(3)]).transpose((1, 2, 3, 0)) # (ng1, ng2, ng3, 3)
            para = np.sum([pa[w]*ra[:, :, :, w] for w in range(3)], axis=0)
            ll2   = pa.dot(pa)*np.sum([ra[:, :, :, w]*ra[:, :, :, w] for w in range(3)], axis=0)
            angle1 = np.arccos(para/(1.0e-9+np.sqrt(ll2)))

            # angle between (a-p) and (r-p):
            ap = a-p # (3)
            rp = np.array([r_gv[:, :, :, w]-p[w] for w in range(3)]).transpose((1, 2, 3, 0)) # (ng1, ng2, ng3, 3)
            aprp = np.sum([ap[w]*rp[:, :, :, w] for w in range(3)], axis=0)
            ll2   = ap.dot(ap)*np.sum([rp[:, :, :, w]*rp[:, :, :, w] for w in range(3)], axis=0)
            angle2 = np.arccos(aprp/(1.0e-9+np.sqrt(ll2)))

            # Include in the mask
            this_mask = np.logical_and(np.logical_and(angle1 < 0.5*np.pi, angle2 < 0.5*np.pi),
                                      d < self.radius**2.0 )

            # Add spheres around current end points
            if self.round_corners:
                # |r-a| and |r-p|
                raDist = np.sum([ra[:, :, :, w]*ra[:, :, :, w] for w in range(3)], axis=0)
                rpDist = np.sum([rp[:, :, :, w]*rp[:, :, :, w] for w in range(3)], axis=0)
                this_mask = np.logical_or(this_mask,
                                         np.logical_or(raDist < self.radius**2.0, rpDist < self.radius**2.0))

            mask =  np.logical_or(mask, this_mask)

            # move to next point
            a = p

        return mask

class PolarizableTetrahedron():
    #http://steve.hollasch.net/cgindex/geometry/ptintet.html
    #obrecht@imagen.com (Doug Obrecht) writes:
    #
    # Can someone point me to an algorithm that determines if a point is within a tetrahedron?
    #
    # Let the tetrahedron have vertices
    #     V1 = (x1, y1, z1)
    #    V2 = (x2, y2, z2)
    #    V3 = (x3, y3, z3)
    #    V4 = (x4, y4, z4)
    #
    #and your test point be
    #
    #        P = (x, y, z).
    #Then the point P is in the tetrahedron if following five determinants all have the same sign.
    #
    #             |x1 y1 z1 1|
    #        D0 = |x2 y2 z2 1|
    #             |x3 y3 z3 1|
    #             |x4 y4 z4 1|
    #
    #             |x  y  z  1|
    #        D1 = |x2 y2 z2 1|
    #             |x3 y3 z3 1|
    #             |x4 y4 z4 1|
    #
    #             |x1 y1 z1 1|
    #        D2 = |x  y  z  1|
    #             |x3 y3 z3 1|
    #             |x4 y4 z4 1|
    #
    #             |x1 y1 z1 1|
    #        D3 = |x2 y2 z2 1|
    #             |x  y  z  1|
    #             |x4 y4 z4 1|
    #
    #             |x1 y1 z1 1|
    #        D4 = |x2 y2 z2 1|
    #             |x3 y3 z3 1|
    #             |x  y  z  1|
    #
    # Some additional notes:
    #
    # If by chance the D0=0, then your tetrahedron is degenerate (the points are coplanar).
    # If any other Di=0, then P lies on boundary i (boundary i being that boundary formed by the three points other than Vi).
    # If the sign of any Di differs from that of D0 then P is outside boundary i.
    # If the sign of any Di equals that of D0 then P is inside boundary i.
    # If P is inside all 4 boundaries, then it is inside the tetrahedron.
    # As a check, it must be that D0 = D1+D2+D3+D4.
    # The pattern here should be clear; the computations can be extended to simplicies of any dimension. (The 2D and 3D case are the triangle and the tetrahedron).
    # If it is meaningful to you, the quantities bi = Di/D0 are the usual barycentric coordinates.
    # Comparing signs of Di and D0 is only a check that P and Vi are on the same side of boundary i.

    def __init__(self, *, corners, permittivity):
        # sanity check
        assert(len(corners)==4)     # exactly 4 points
        assert(len(corners[0])==3)  # 3D

        self.name = 'PolarizableTetrahedron'
        self.arguments = 'corners=['
        for c in corners:
            self.arguments += '[%20.12e, %20.12e, %20.12e],' % (c[0], c[1], c[2])
        self.arguments += ']'

        self.corners      = np.array(corners)/Bohr # from Angstroms to atomic units
        self.permittivity = permittivity

    def determinant_value(self, x, y, z,
                       x1, y1, z1,
                       x2, y2, z2,
                       x3, y3, z3,
                       x4, y4, z4, ind):
        mat = np.array([[x1, y1, z1, 1], [x2, y2, z2, 1], [x3, y3, z3, 1], [x4, y4, z4, 1]])
        mat[ind][:] = np.array([x, y, z, 1])
        return np.linalg.det(mat)

    def get_mask(self, gd):
        r_gv = gd.get_grid_point_coordinates().transpose((1, 2, 3, 0))
        ng   = r_gv.shape[0:-1]
        ngv  = r_gv.shape
        x1, y1, z1 = self.corners[0]
        x2, y2, z2 = self.corners[1]
        x3, y3, z3 = self.corners[2]
        x4, y4, z4 = self.corners[3]

        mask = np.ones(ng)==0

        # TODO: associate a determinant for each point, and use numpy tools to determine
        #       the mask without the *very slow* loop over grid points
        for ind, pt in np.ndenumerate(r_gv[:, :, :, 0]):
            x, y, z = r_gv[ind][:]
            d0 = np.array([[x1, y1, z1, 1],
                           [x2, y2, z2, 1],
                           [x3, y3, z3, 1],
                           [x4, y4, z4, 1]])
            d1 = np.array([[x,  y,  z, 1],
                           [x2, y2, z2, 1],
                           [x3, y3, z3, 1],
                           [x4, y4, z4, 1]])
            d2 = np.array([[x1, y1, z1, 1],
                           [x,  y,  z,  1],
                           [x3, y3, z3, 1],
                           [x4, y4, z4, 1]])
            d3 = np.array([[x1, y1, z1, 1],
                           [x2, y2, z2, 1],
                           [x,  y,  z,  1],
                           [x4, y4, z4, 1]])
            d4 = np.array([[x1, y1, z1, 1],
                           [x2, y2, z2, 1],
                           [x3, y3, z3, 1],
                           [x,  y,  z,  1]])
            s0 = np.linalg.det(d0)
            s1 = np.linalg.det(d1)
            s2 = np.linalg.det(d2)
            s3 = np.linalg.det(d3)
            s4 = np.linalg.det(d4)

            if (np.sign(s0) == np.sign(s1) or abs(s1) < 1e-12) and \
               (np.sign(s0) == np.sign(s2) or abs(s2) < 1e-12) and \
               (np.sign(s0) == np.sign(s3) or abs(s3) < 1e-12) and \
               (np.sign(s0) == np.sign(s4) or abs(s4) < 1e-12):
                mask[ind] = True
        return mask

# Lorentzian oscillator function: L(omega) = eps0 * beta / (w**2 - i*alpha*omega - omega**2)    // Coomar2011, Eq. 2
class LorentzOscillator:
    def __init__(self, bar_omega, alpha, beta):
        self.bar_omega = bar_omega
        self.alpha    = alpha
        self.beta     = beta

    def value(self, omega):
        return _eps0_au * self.beta / (self.bar_omega**2 - 1J * self.alpha * omega - omega**2)

# Dieletric function: e(omega) = eps_inf + sum_j L_j(omega) // Coomar2011, Eq. 2
class Permittivity:
    def __init__(self, fname=None, data=None, eps_infty = _eps0_au ):

        # Initialize to vacuum permittivity
        self.eps_infty = eps_infty
        self.Nj = 0
        self.oscillators = []

        # Input as data array
        if data is not None:
            assert fname is None
            self.Nj = len(data)
            for v in range(self.Nj):
                self.oscillators.append(LorentzOscillator(data[v][0]/Hartree, data[v][1]/Hartree, data[v][2]/Hartree/Hartree))
            return

        # Input as filename
        if fname != None: # read permittivity from a 3-column file
            with open(fname, 'r') as fp:
                lines = fp.readlines()

            self.Nj = len(lines)

            for line in lines:
                bar_omega = float(line.split()[0]) / Hartree
                alpha     = float(line.split()[1]) / Hartree
                beta      = float(line.split()[2]) / Hartree / Hartree
                self.oscillators.append(LorentzOscillator(bar_omega, alpha, beta))
            return



    def value(self, omega = 0):
        return self.eps_infty + sum([osc.value(omega) for osc in self.oscillators])

    def data(self):
        return [[osc.bar_omega, osc.alpha, osc.beta] for osc in self.oscillators]

    def data_eVA(self):
        return [[osc.bar_omega*Hartree, osc.alpha*Hartree, osc.beta*Hartree*Hartree] for osc in self.oscillators]

# Dieletric function that renormalizes the static permittivity to the requested value (usually epsZero)
class PermittivityPlus(Permittivity):
    def __init__(self, fname=None, data=None, eps_infty = _eps0_au, epsZero = _eps0_au, newbar_omega = 0.01, new_alpha = 0.10, **kwargs):
        Permittivity.__init__(self, fname=fname, data=data, eps_infty=eps_infty)

        # Convert given values from eVs to Hartrees
        _newbar_omega = newbar_omega / Hartree
        _new_alpha    = new_alpha / Hartree

        # Evaluate the new value
        _new_beta = ((epsZero - self.value(0.0))*_newbar_omega**2.0/_eps0_au).real
        self.oscillators.append(LorentzOscillator(_newbar_omega, _new_alpha, _new_beta))
        self.Nj = len(self.oscillators)




import numpy as np
import pytest
from ase.build import fcc111

from gpaw import GPAW, LCAO, FermiDirac
from gpaw.mpi import world
from gpaw.utilities import compiled_with_sl

# This test verifies that energy and forces are (approximately)
# parallelization independent
#
# Tests the LCAO energy and forces in non-orthogonal cell with
# simultaneous parallelization over bands, domains and k-points (if
# enough CPUs are available), both with and without scalapack
# (if scalapack is available).
#
# Run with 1, 2, 4 or 8 (best) CPUs.
#
# This test covers many cases not caught by lcao_parallel or
# lcao_parallel_kpt
#
# Written November 24, 2011, r8567


@pytest.mark.later
def test_lcao_complicated():
    system = fcc111('Au', size=(1, 3, 1))
    system.numbers[0] = 8
    # It is important that the number of atoms is uneven; this
    # tests the case where the band parallelization does not match
    # the partitioning of orbitals between atoms (the middle atom has orbitals
    # on distinct band descriptor ranks)

    system.center(vacuum=3.5, axis=2)
    system.rattle(stdev=0.2, seed=17)
    # from ase.visualize import view
    # view(system)

    # system.set_pbc(0)
    # system.center(vacuum=3.5)

    def calculate(parallel, comm=world, Eref=None, Fref=None):
        calc = GPAW(mode=LCAO(atomic_correction='sparse'),
                    basis=dict(O='dzp', Au='sz(dzp)'),
                    occupations=FermiDirac(0.1),
                    kpts=(4, 1, 1),
                    # txt=None,
                    communicator=comm,
                    nbands=16,
                    parallel=parallel,
                    h=0.35)
        system.calc = calc
        E = system.get_potential_energy()
        F = system.get_forces()

        if world.rank == 0:
            print('Results')
            print('-----------')
            print(E)
            print(F)
            print('-----------')

        if Eref is not None:
            Eerr = abs(E - Eref)
            assert Eerr < 1e-8, 'Bad E: err=%f; parallel=%s' % (Eerr, parallel)
        if Fref is not None:
            Ferr = np.abs(F - Fref).max()
            assert Ferr < 1e-6, 'Bad F: err=%f; parallel=%s' % (Ferr, parallel)
        return E, F

    # First calculate reference energy and forces E and F
    #
    # If we want to really dumb things down, enable this to force an
    # entirely serial calculation:
    if 0:
        serial = world.new_communicator([0])
        E = 0.0
        F = np.zeros((len(system), 3))
        if world.rank == 0:
            E, F = calculate({}, serial)
        E = world.sum(E)
        world.sum(F)
    else:
        # Normally we'll just do it in parallel;
        # that case is covered well by other tests, so we can probably trust it
        E, F = calculate({}, world)

    def check(parallel):
        return calculate(parallel, comm=world, Eref=E, Fref=F)

    assert world.size in [1, 2, 4, 8], ('Number of CPUs %d not supported'
                                        % world.size)

    parallel = dict(domain=1, band=1)
    if world.size % 2 == 0:
        parallel['band'] = 2
    if world.size % 4 == 0:
        parallel['domain'] = 2

    # If size is 8, this will also use kpt parallelization.  This test should
    # run with 8 CPUs for best coverage of parallelizations
    if world.size == 8:
        pass  # sl_cpus = 4 ???

    if world.size > 1:
        check(parallel)

    if compiled_with_sl() and world.size > 1:
        parallel['sl_auto'] = True
        check(parallel)



"""
Test the dipole correction code by comparing this system:

    H
z1   O  z2
    H


(where z1 and z2 denote points where the potential is probed)

Expected potential:

      -----
     /
    /
----

to this system:

   H           H
z1  O    z2   O
   H           H


Expected potential:

       -------
      /       \
     /         \
-----           ------

The height of the two potentials are tested to be the same.

Enable if-statement in the bottom for nice plots
"""
import numpy as np
from ase.build import molecule
from ase.units import Hartree
from gpaw import GPAW, Mixer
from gpaw.mpi import rank


def test_dipole():
    system1 = molecule('H2O')
    system1.set_pbc((True, True, False))
    system1.cell = 4.0 * np.array([[1.0, -1.5, 0.0], [1.0, 1.0, 0.0],
                                   [0., 0., 1.]])
    system1.center(vacuum=10.0, axis=2)

    system2 = system1.copy()
    system2.positions *= [1.0, 1.0, -1.0]
    system2 += system1
    system2.center(vacuum=6.0, axis=2)

    convergence = dict(density=1e-6)

    def kw():
        return dict(mode='lcao',
                    convergence=convergence,
                    mixer=Mixer(0.5, 7, 50.0),
                    h=0.25,
                    xc='oldLDA')

    calc1 = GPAW(poissonsolver={'dipolelayer': 'xy'}, **kw())

    system1.calc = calc1
    system1.get_potential_energy()
    v1 = calc1.get_effective_potential()

    calc2 = GPAW(**kw())

    system2.calc = calc2
    system2.get_potential_energy()
    v2 = calc2.get_effective_potential()

    def get_avg(v):
        nx, ny, nz = v.shape
        vyz = v.sum(axis=0) / nx
        vz = vyz.sum(axis=0) / ny
        return vz, vyz

    if rank == 0:
        vz1, vyz1 = get_avg(v1)
        vz2, vyz2 = get_avg(v2)

        # Compare values that are not quite at the end of the array
        # (at the end of the array things can "oscillate" a bit)
        dvz1 = vz1[-3] - vz1[3]
        dvz2 = vz2[3] - vz2[len(vz2) // 2]
        print(dvz1, dvz2)

        err1 = abs(dvz1 - dvz2)

        try:
            correction = calc1.hamiltonian.poisson.correction
        except AttributeError:
            correction = (calc1.calculation.pot_calc.poisson_solver.
                          solver.correction)

        correction_err = abs(2.0 * correction * Hartree + dvz1)
        print('correction error %s' % correction_err)
        assert correction_err < 3e-5

        # Comparison to what the values were when this test was last modified:
        ref_value = 2.07342988218
        err2 = abs(dvz1 - ref_value)

        if 0:
            import matplotlib.pyplot as plt
            plt.imshow(vyz1)
            plt.figure()
            plt.imshow(vyz2)
            plt.figure()
            plt.plot(vz1)
            plt.plot(vz2)
            plt.show()

        print('Ref value of previous calculation', ref_value)
        print('Value in this calculation', dvz1)

        # fine grid needed to achieve convergence!
        print('Error', err1, err2)
        assert err1 < 4e-3, err1
        assert err2 < 2e-4, err2



import numpy as np

from gpaw.mpi import world
from gpaw.blacs import BlacsGrid, Redistributor


def check(comm, M, N, mcpus, ncpus, mb, nb):
    grid0 = BlacsGrid(comm, 1, 1)
    desc0 = grid0.new_descriptor(M, N, M, N, 0, 0)
    A_mn = desc0.zeros(dtype=float)
    A_mn[:] = comm.size + 1

    grid1 = BlacsGrid(comm, mcpus, ncpus)
    desc1 = grid1.new_descriptor(M, N, mb, nb, 0, 0)  # ???
    B_mn = desc1.zeros(dtype=float)
    B_mn[:] = comm.rank

    if comm.rank == 0:
        msg = 'Slices of global matrix indices by rank'
        print(msg)
        print('-' * len(msg))

    for rank in range(comm.size):
        comm.barrier()
        if rank == comm.rank:
            print('Rank %d:' % rank)
            last_Mstart = -1
            for Mstart, Mstop, Nstart, Nstop, block in desc1.my_blocks(B_mn):
                if Mstart > last_Mstart and last_Mstart >= 0:
                    print()
                print('[%3d:%3d, %3d:%3d]' % (Mstart, Mstop, Nstart, Nstop),
                      end=' ')
                last_Mstart = Mstart
                assert (block == comm.rank).all()
            print()
            print()
        comm.barrier()

    redistributor = Redistributor(comm, desc1, desc0)
    redistributor.redistribute(B_mn, A_mn)

    if comm.rank == 0:
        msg = 'Rank where each element of the global matrix is stored'
        print(msg)
        print('-' * len(msg))
        print(A_mn)


def test_parallel_blacsdist(scalapack):
    M, N = 10, 10
    mb, nb = 2, 2
    mcpus = int(np.ceil(world.size**0.5))
    ncpus = world.size // mcpus

    if world.rank == 0:
        print('world size:   ', world.size)
        print('M x N:        ', M, 'x', N)
        print('mcpus x ncpus:', mcpus, 'x', ncpus)
        print('mb x nb:      ', mb, 'x', nb)
        print()

    check(world, M, N, mcpus, ncpus, mb, nb)



import numpy as np
import pytest
from ase.build import bulk
from ase.parallel import parprint

from gpaw import GPAW, PW, Mixer
from gpaw.mpi import world


@pytest.mark.mgga
def test_pw_si_stress_mgga(in_tmp_dir):
    xc = 'TPSS'
    si = bulk('Si')
    k = 3
    si.calc = GPAW(mode=PW(250),
                   mixer=Mixer(0.7, 5, 50.0),
                   xc=xc,
                   kpts=(k, k, k),
                   convergence={'energy': 1e-8},
                   parallel={'domain': min(2, world.size)},
                   txt='si.txt')

    si.set_cell(np.dot(si.cell,
                       [[1.02, 0, 0.03],
                        [0, 0.99, -0.02],
                        [0.2, -0.01, 1.03]]),
                scale_atoms=True)

    si.get_potential_energy()

    # Trigger nasty bug (fixed in !486):
    si.calc.wfs.pt.blocksize = si.calc.wfs.pd.maxmyng - 1

    s_analytical = si.get_stress()
    # Calculated numerical stress once, store here to speed up test
    # [-0.00853353 -0.04100137 -0.03821226 -0.02084923 0.1331957 0.00945546]
    s_numerical = np.array([-0.00853353, -0.04100137, -0.03821226,
                            -0.02084923, 0.1331957, 0.00945546])
    # s_numerical = si.calc.calculate_numerical_stress(si, 1e-5)
    s_err = s_numerical - s_analytical

    parprint('Analytical stress:\n', s_analytical)
    parprint('Numerical stress:\n', s_numerical)
    parprint('Error in stress:\n', s_err)
    assert np.all(abs(s_err) < 1e-4)



import pytest
import numpy as np
from gpaw.test import equal
from gpaw.grid_descriptor import GridDescriptor
from gpaw.pw.descriptor import PWDescriptor
from gpaw.mpi import world


@pytest.mark.ci
def test_pw_interpol():
    def test(gd1, gd2, pd1, pd2, R1, R2):
        a1 = gd1.zeros(dtype=pd1.dtype)
        a1[R1] = 1
        a2 = pd1.interpolate(a1, pd2)[0]
        x = a2[R2]

        a2 = gd2.zeros(dtype=pd2.dtype)
        a2[R2] = 1
        y = pd2.restrict(a2, pd1)[0][R1] * a2.size / a1.size

        equal(x, y, 1e-9)
        return x

    if world.size == 1:
        for size1, size2 in [
            [(3, 3, 3), (8, 8, 8)],
            [(4, 4, 4), (9, 9, 9)],
            [(2, 4, 4), (5, 9, 9)],
            [(2, 3, 4), (5, 6, 9)],
            [(2, 3, 4), (5, 6, 8)],
            [(4, 4, 4), (8, 8, 8)],
            [(2, 4, 4), (4, 8, 8)],
            [(2, 4, 2), (4, 8, 4)]]:
            print(size1, size2)
            gd1 = GridDescriptor(size1, size1)
            gd2 = GridDescriptor(size2, size1)
            pd1 = PWDescriptor(1, gd1, complex)
            pd2 = PWDescriptor(1, gd2, complex)
            pd1r = PWDescriptor(1, gd1)
            pd2r = PWDescriptor(1, gd2)
            for R1, R2 in [[(0, 0, 0), (0, 0, 0)],
                           [(0, 0, 0), (0, 0, 1)]]:
                x = test(gd1, gd2, pd1, pd2, R1, R2)
                y = test(gd1, gd2, pd1r, pd2r, R1, R2)
                equal(x, y, 1e-9)

            a1 = np.random.random(size1)
            a2 = pd1r.interpolate(a1, pd2r)[0]
            c2 = pd1.interpolate(a1 + 0.0j, pd2)[0]
            d2 = pd1.interpolate(a1 * 1.0j, pd2)[0]
            equal(abs(c2.imag).max(), 0, 1e-14)
            equal(abs(d2.real).max(), 0, 1e-14)
            equal(gd1.integrate(a1), gd2.integrate(a2), 1e-13)
            equal(abs(c2 - a2).max(), 0, 1e-14)
            equal(abs(d2 - a2 * 1.0j).max(), 0, 1e-14)

            a1 = pd2r.restrict(a2, pd1r)[0]
            c1 = pd2.restrict(a2 + 0.0j, pd1)[0]
            d1 = pd2.restrict(a2 * 1.0j, pd1)[0]
            equal(gd1.integrate(a1), gd2.integrate(a2), 1e-13)
            equal(abs(c1 - a1).max(), 0, 1e-14)
            equal(abs(d1 - a1 * 1.0j).max(), 0, 1e-14)



from math import sin, cos, pi, sqrt

import numpy as np

from ase.units import Bohr, Hartree, alpha
from gpaw.fd_operators import Gradient
from gpaw.utilities.gl_quadrature import GaussLegendre
from gpaw.pes import ds_prefactor
from gpaw.pes.state import H1s
from gpaw.pes.continuum import PlaneWave

# debug
# from gpaw.mpi import rank


class CrossSectionBeta:

    def __init__(self,
                 initial=None,
                 final=None,
                 r0=[0, 0, 0],  # center of mass vector
                 form='L',
                 ngauss=8):

        self.initial = initial
        self.r0 = np.array(r0) / Bohr
        self.form = form
        if final is None:
            self.final = PlaneWave(initial.gd)
        else:
            self.final = final

        self.Ekin = None

        # Gauss-Legendre weights and abscissas
        self.gl = {}
        self.gl['x'] = GaussLegendre(-1., 1., ngauss)
        self.gl['phi'] = GaussLegendre(0, 2 * pi, ngauss)
        self.gl['psi'] = self.gl['phi']
        self.angle = {}

        # sin and cos of the magic angle (54.7 deg)
        self.costhm = 1. / sqrt(3)
        self.sinthm = sqrt(2. / 3.)

    def calculate(self, Ekin):
        """Calculate the necessary overlaps."""

        Ekin = Ekin / Hartree
        if self.Ekin == Ekin:
            return
        self.Ekin = Ekin

        # photoelectron momentum
        self.k = sqrt(2 * self.Ekin)

        for angle in ['x', 'phi', 'psi']:
            self.angle[angle] = self.gl[angle].get_x()[0]
        self.T20, self.T2m = self.gi_x()

        # we need the average
        self.T20 /= 8 * pi ** 2
        self.T2m /= 8 * pi ** 2

    def get_omega(self):
        """Return the necessary photon energy."""
        return self.Ekin - self.initial.get_energy() / Hartree

    def get_beta(self, Ekin=None):
        """Return the asymmetry parameter.

        E: photoelectron kinetic energy [eV]
        """
        if Ekin is not None:
            self.calculate(Ekin)
        return self.T20 / self.T2m - 1.

    def get_ds(self, Ekin=None, units='Mb'):
        """Return the total cross section.

        Ekin: photoelectron kinetic energy [eV]
        units:
        'Mb', 1 Mb = 1.e-22 m**2
        'Ang', 1 A**2 = 1.e-20 m**2
        'a.u.', 1 a_0**2 = 2.8e-21 m**2
        as output units
        """
        if Ekin is not None:
            self.calculate(Ekin)
        try:
            pre = ds_prefactor[units]
        except KeyError:
            print('Unknown units: >' + units + '<')
            raise

#        me_c =  self.initial.get_me_c(np.array([0., 0., self.k]), self.form)
#        T2mana = np.abs(np.dot(me_c,me_c)) / 3.
#        print "T2m:", T2mana, self.T2m

        omega = self.get_omega()

        # integration over momentum agles
        pre *= self.k * 4 * pi

#        print omega, self.initial.get_ds(self.k, omega, self.form), \
#            (self.k * 4 * pi * (2 * pi)**2 / 137.0359895 * self.T2m / omega)

        return pre * ((2 * pi) ** 2 * alpha * self.T2m / omega)

    def gauss_integrate(self, angle, function):
        T20 = 0.
        T2m = 0.
        gl = self.gl[angle]
        for x, w in zip(gl.get_x(), gl.get_w()):
            self.angle[angle] = x
            t20, t2m = function()
            T20 += w * t20
            T2m += w * t2m
#        print "<gauss_integrate> angle=", angle, T2m
        return T20, T2m

    def gi_x(self):
        """Gauss integrate over x=cos(theta)"""
        return self.gauss_integrate('x', self.gi_phi)

    def gi_phi(self):
        """Gauss integrate over phi"""
        return self.gauss_integrate('phi', self.gi_psi)

    def gi_psi(self):
        """Gauss integrate over psi"""
        return self.gauss_integrate('psi', self.integrand)

    def integrand(self):

        # polarisation in the direction of vk
        costh = self.angle['x']
        sinth = sqrt(1. - costh ** 2)
        sinphi = sin(self.angle['phi'])
        cosphi = cos(self.angle['phi'])
        eps0 = np.array([sinth * cosphi,
                         sinth * sinphi,
                         costh])
        vk = self.k * eps0

        # polarisation at the magic angle
        costhm = self.costhm
        sinthm = self.sinthm
        sinpsi = sin(self.angle['psi'])
        cospsi = cos(self.angle['psi'])
        epsm = np.array([sinthm * (cosphi * sinpsi * costh +
                                   sinphi * cospsi) +
                         costhm * cosphi * sinth,
                         sinthm * (sinphi * sinpsi * costh -
                                   cosphi * cospsi) +
                         costhm * sinphi * sinth,
                         costhm * costh - sinthm * sinth * sinpsi])

        # initial and final state on the grid
        initial_G = self.initial.get_grid()
        final_G = self.final.get_grid(vk, self.r0)
        ini_analyt = H1s(self.initial.gd, self.r0)

        gd = self.initial.gd
        if self.form == 'L':
            if_G = initial_G * final_G
            omega = self.get_omega()
            if 0:
                me_c = []
                for c in range(3):
                    xyz_G = ((np.arange(gd.n_c[c]) + gd.beg_c[c]) * gd.h_c[c]
                             - self.r0[c])
                    shape = [1, 1, 1]
                    shape[c] = -1
                    xyz_G.shape = tuple(shape)
                    np.resize(xyz_G, gd.n_c)
                    me_c.append(gd.integrate(if_G * xyz_G))
                me_c = np.array(me_c) * omega
            else:
                me_c = gd.calculate_dipole_moment(if_G)
                me_c += self.r0 * gd.integrate(if_G)
                me_c *= -omega
        elif self.form == 'V':
            dtype = final_G.dtype
            phase_cd = np.ones((3, 2), dtype)
            if not hasattr(gd, 'ddr'):
                gd.ddr = [Gradient(gd, c, dtype=dtype).apply for c in range(3)]
            dfinal_G = gd.empty(dtype=dtype)
            me_c = np.empty(3, dtype=dtype)
            for c in range(3):
                gd.ddr[c](final_G, dfinal_G, phase_cd)
                me_c[c] = gd.integrate(initial_G * dfinal_G)
        else:
            raise NotImplementedError

        if 0:
            omega = self.get_omega()
            me_analyt = ini_analyt.get_me_c(vk, self.form)[0].imag
            me = me_c[0].imag

            def ds(me):
                return self.k / omega * me ** 2

            print(omega, ds(me_analyt), ds(me), me_analyt, me)
#        print 'analyt', self.initial.get_me_c(vk, self.form)
#        print 'num', me_c
#        print 'analyt/num', self.initial.get_me_c(vk, self.form) / me_c

        # return the squared matrix elements
        T2 = []
        for eps in [eps0, epsm]:
            me = np.dot(eps, me_c)
#            print "eps, T2:", eps, (me * me.conj()).real
            T2.append((me * me.conj()).real)
#        print vk, T2
        return T2[0], T2[1]



from __future__ import annotations

import numbers
from typing import Any

import numpy as np
from ase.units import Ha
from gpaw.typing import ArrayLike1D


class FrequencyDescriptor:
    def __init__(self, omega_w: ArrayLike1D):
        """Frequency grid descriptor.

        Parameters
        ----------
        omega_w:
            Frequency grid in Hartree units.
        """
        self.omega_w = np.asarray(omega_w).copy()

    def __len__(self):
        return len(self.omega_w)

    def __repr__(self):
        emin = self.omega_w[0] * Ha
        emax = self.omega_w[-1] * Ha
        return (f'{self.__class__.__name__}'
                f'(from {emin:.3f} to {emax:.3f} eV, {len(self)} points)')

    @staticmethod
    def from_array_or_dict(input: dict[str, Any] | ArrayLike1D
                           ) -> FrequencyDescriptor:
        """Create frequency-grid descriptor.

        In case *input* is a list on frequencies (in eV) a
        :class:`FrequencyGridDescriptor` instance is returned.
        Othervise a :class:`NonLinearFrequencyDescriptor` instance is
        returned.

        >>> from ase.units import Ha
        >>> params = dict(type='nonlinear',
        ...               domega0=0.1,
        ...               omega2=10,
        ...               omegamax=50)
        >>> wd = FrequencyDescriptor.from_array_or_dict(params)
        >>> wd.omega_w[0:2] * Ha
        array([0.        , 0.10041594])
        """
        if isinstance(input, dict):
            assert input['type'] == 'nonlinear'
            domega0 = input.get('domega0')
            omega2 = input.get('omega2')
            omegamax = input['omegamax']
            return NonLinearFrequencyDescriptor(
                (0.1 if domega0 is None else domega0) / Ha,
                (10.0 if omega2 is None else omega2) / Ha,
                omegamax / Ha)
        return FrequencyGridDescriptor(np.asarray(input) / Ha)


class ComplexFrequencyDescriptor:

    def __init__(self, hz_z: ArrayLike1D):
        """Construct the complex frequency descriptor.

        Parameters
        ----------
        hz_z:
            Array of complex frequencies (in units of Hartree)
        """
        # Use a copy of the input array
        hz_z = np.array(hz_z)
        assert hz_z.dtype == complex

        self.hz_z = hz_z

    def __len__(self):
        return len(self.hz_z)

    def almost_eq(self, zd):
        if len(zd) != len(self):
            return False
        return np.allclose(self.hz_z, zd.hz_z)

    @staticmethod
    def from_array(frequencies: ArrayLike1D):
        """Create a ComplexFrequencyDescriptor from frequencies in eV."""
        return ComplexFrequencyDescriptor(np.asarray(frequencies) / Ha)

    @property
    def upper_half_plane(self):
        """All frequencies reside in the upper half complex frequency plane?"""
        return np.all(self.hz_z.imag > 0.)

    @property
    def horizontal_contour(self):
        """Do all frequency point lie on a horizontal contour?"""
        return np.ptp(self.hz_z.imag) < 1.e-6

    @property
    def omega_w(self):
        """Real part of the frequencies."""
        assert self.horizontal_contour,\
            'It only makes sense to index the frequencies by their real part '\
            'if they reside on a horizontal contour.'
        return self.hz_z.real


class FrequencyGridDescriptor(FrequencyDescriptor):

    def get_index_range(self, lim1_m, lim2_m):
        """Get index range. """

        i0_m = np.zeros(len(lim1_m), int)
        i1_m = np.zeros(len(lim2_m), int)

        for m, (lim1, lim2) in enumerate(zip(lim1_m, lim2_m)):
            i_x = np.logical_and(lim1 <= self.omega_w,
                                 lim2 >= self.omega_w)
            if i_x.any():
                inds = np.argwhere(i_x)
                i0_m[m] = inds.min()
                i1_m[m] = inds.max() + 1

        return i0_m, i1_m


class NonLinearFrequencyDescriptor(FrequencyDescriptor):
    def __init__(self,
                 domega0: float,
                 omega2: float,
                 omegamax: float):
        """Non-linear frequency grid.

        Units are Hartree.  See :ref:`frequency grid`.

        Parameters
        ----------
        domega0:
            Frequency grid spacing for non-linear frequency grid at omega = 0.
        omega2:
            Frequency at which the non-linear frequency grid has doubled
            the spacing.
        omegamax:
            The upper frequency bound for the non-linear frequency grid.
        """
        beta = (2**0.5 - 1) * domega0 / omega2
        wmax = int(omegamax / (domega0 + beta * omegamax))
        w = np.arange(wmax + 2)  # + 2 is for buffer
        omega_w = w * domega0 / (1 - beta * w)

        super().__init__(omega_w)

        self.domega0 = domega0
        self.omega2 = omega2
        self.omegamax = omegamax
        self.omegamin = 0

        self.beta = beta
        self.wmax = wmax
        self.omega_w = omega_w
        self.wmax = wmax

    def get_floor_index(self, o_m, safe=True):
        """Get closest index rounding down."""
        beta = self.beta
        w_m = (o_m / (self.domega0 + beta * o_m)).astype(int)
        if safe:
            if isinstance(w_m, np.ndarray):
                w_m[w_m >= self.wmax] = self.wmax - 1
            elif isinstance(w_m, numbers.Integral):
                if w_m >= self.wmax:
                    w_m = self.wmax - 1
            else:
                raise TypeError
        return w_m

    def get_index_range(self, omega1_m, omega2_m):
        omega1_m = omega1_m.copy()
        omega2_m = omega2_m.copy()
        omega1_m[omega1_m < 0] = 0
        omega2_m[omega2_m < 0] = 0
        w1_m = self.get_floor_index(omega1_m)
        w2_m = self.get_floor_index(omega2_m)
        o1_m = self.omega_w[w1_m]
        o2_m = self.omega_w[w2_m]
        w1_m[o1_m < omega1_m] += 1
        w2_m[o2_m < omega2_m] += 1
        return w1_m, w2_m



import numpy as np


def get_goldstone_scaling(mode, chiks, Kxc_GG, dyson_solver):
    """Get kernel scaling parameter to fulfill a Goldstone condition."""
    assert chiks.qpd.kd.gamma,\
        r'The Goldstone condition is strictly bound to the Γ-point'

    # Find the frequency to determine the scaling from
    omega_w = chiks.zd.omega_w
    wgs = find_goldstone_frequency(mode, omega_w)

    # Only one rank, rgs, has the given frequency
    assert chiks.distribution == 'zGG',\
        'Only block distribution over frequencies is allowed at this point'
    wblocks = chiks.blocks1d
    rgs, mywgs = wblocks.find_global_index(wgs)

    # Let rgs find the rescaling
    fxcsbuf = np.empty(1, dtype=float)
    if wblocks.blockcomm.rank == rgs:
        chiks_GG = chiks.array[mywgs]
        fxcsbuf[:] = find_goldstone_scaling(mode, chiks_GG, Kxc_GG,
                                            dyson_solver)

    # Broadcast found rescaling
    wblocks.blockcomm.broadcast(fxcsbuf, rgs)
    fxcs = fxcsbuf[0]

    return fxcs


def find_goldstone_frequency(mode, omega_w):
    """Factory function for finding the appropriate frequency to determine
    the kernel scaling from according to different Goldstone criteria."""
    if mode == 'fm':
        return find_fm_goldstone_frequency(omega_w)
    elif mode == 'afm':
        return find_afm_goldstone_frequency(omega_w)
    else:
        raise ValueError(
            f"Allowed Goldstone scaling modes are 'fm', 'afm'. Got: {mode}")


def find_fm_goldstone_frequency(omega_w):
    """Find omega=0. as the fm Goldstone frequency."""
    wgs = np.abs(omega_w).argmin()
    if not np.allclose(omega_w[wgs], 0., atol=1.e-8):
        raise ValueError("Frequency grid needs to include"
                         + " omega=0. to allow 'fm' Goldstone scaling")

    return wgs


def find_afm_goldstone_frequency(omega_w):
    """Find the second smallest positive frequency
    as the afm Goldstone frequency."""
    # Set omega=0. and negative frequencies to np.inf
    omega1_w = np.where(omega_w < 1.e-8, np.inf, omega_w)
    # Sort for the two smallest positive frequencies
    omega2_w = np.partition(omega1_w, 1)
    # Find original index of second smallest positive frequency
    wgs = np.abs(omega_w - omega2_w[1]).argmin()

    return wgs


def find_goldstone_scaling(mode, chiks_GG, Kxc_GG, dyson_solver):
    """Factory function for finding the scaling of the kernel
    according to different Goldstone criteria."""
    assert mode in ['fm', 'afm'],\
        f"Allowed Goldstone scaling modes are 'fm', 'afm'. Got: {mode}"

    if mode == 'fm':
        return find_fm_goldstone_scaling(chiks_GG, Kxc_GG, dyson_solver)
    elif mode == 'afm':
        return find_afm_goldstone_scaling(chiks_GG, Kxc_GG, dyson_solver)


def find_fm_goldstone_scaling(chiks_GG, Kxc_GG, dyson_solver):
    """Find goldstone scaling of the kernel by ensuring that the
    macroscopic inverse enhancement function has a root in (q=0, omega=0)."""
    fxcs = 1.
    kappaM = calculate_macroscopic_kappa(chiks_GG, Kxc_GG * fxcs, dyson_solver)
    # If kappaM > 0, increase scaling (recall: kappaM ~ 1 - Kxc Re{chi_0})
    scaling_incr = 0.1 * np.sign(kappaM)
    while abs(kappaM) > 1.e-7 and abs(scaling_incr) > 1.e-7:
        fxcs += scaling_incr
        if fxcs <= 0.0 or fxcs >= 10.:
            raise Exception('Found an invalid fxc_scaling of %.4f' % fxcs)

        kappaM = calculate_macroscopic_kappa(
            chiks_GG, Kxc_GG * fxcs, dyson_solver)

        # If kappaM changes sign, change sign and refine increment
        if np.sign(kappaM) != np.sign(scaling_incr):
            scaling_incr *= -0.2

    return fxcs


def find_afm_goldstone_scaling(chiks_GG, Kxc_GG, dyson_solver):
    """Find goldstone scaling of the kernel by ensuring that the
    macroscopic magnon spectrum vanishes at q=0. for finite frequencies."""
    fxcs = 1.
    SM = calculate_macroscopic_spectrum(chiks_GG, Kxc_GG * fxcs, dyson_solver)
    # If SM > 0., increase the scaling. If SM < 0., decrease the scaling.
    scaling_incr = 0.1 * np.sign(SM)
    while (SM < 0. or SM > 1.e-7) or abs(scaling_incr) > 1.e-7:
        fxcs += scaling_incr
        if fxcs <= 0. or fxcs >= 10.:
            raise Exception('Found an invalid fxc_scaling of %.4f' % fxcs)

        SM = calculate_macroscopic_spectrum(
            chiks_GG, Kxc_GG * fxcs, dyson_solver)

        # If chi changes sign, change sign and refine increment
        if np.sign(SM) != np.sign(scaling_incr):
            scaling_incr *= -0.2

    return fxcs


def calculate_macroscopic_kappa(chiks_GG, Kxc_GG, dyson_solver):
    """Invert dyson equation and calculate the inverse enhancement function."""
    chi_GG = dyson_solver.invert_dyson_single_frequency(chiks_GG, Kxc_GG)

    return (chiks_GG[0, 0] / chi_GG[0, 0]).real


def calculate_macroscopic_spectrum(chiks_GG, Kxc_GG, dyson_solver):
    """Invert dyson equation and extract the macroscopic spectrum."""
    chi_GG = dyson_solver.invert_dyson_single_frequency(chiks_GG, Kxc_GG)

    return - chi_GG[0, 0].imag / np.pi



"""This module defines Coulomb and XC kernels for the response model.
"""

import numpy as np
from ase.dft import monkhorst_pack
from gpaw.response.pair_functions import SingleQPWDescriptor


class CoulombKernel:
    def __init__(self, truncation, gs):
        self.truncation = truncation
        assert self.truncation in {None, '0D', '2D'}
        self._gs = gs

    def description(self):
        if self.truncation is None:
            return 'No Coulomb truncation'
        else:
            return f'{self.truncation} Coulomb truncation'

    def sqrtV(self, qpd, q_v):
        return self.V(qpd, q_v)**0.5

    def V(self, qpd, q_v):
        assert isinstance(qpd, SingleQPWDescriptor)
        return get_coulomb_kernel(
            qpd, self._gs.kd.N_c, q_v=q_v,
            truncation=self.truncation)

    def integrated_kernel(self, qpd, reduced):
        return get_integrated_kernel(
            qpd=qpd, N_c=self._gs.kd.N_c,
            truncation=self.truncation, reduced=reduced)


def get_coulomb_kernel(qpd, N_c, truncation=None, q_v=None):
    """Factory function that calls the specified flavour
    of the Coulomb interaction"""

    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
    if q_v is not None:
        assert qpd.kd.gamma
        qG_Gv += q_v

    if truncation is None:
        if qpd.kd.gamma and q_v is None:
            v_G = np.zeros(len(qpd.G2_qG[0]))
            v_G[0] = 4 * np.pi
            v_G[1:] = 4 * np.pi / (qG_Gv[1:]**2).sum(axis=1)
        else:
            v_G = 4 * np.pi / (qG_Gv**2).sum(axis=1)

    elif truncation == '2D':
        v_G = calculate_2D_truncated_coulomb(qpd, q_v=q_v, N_c=N_c)
        if qpd.kd.gamma and q_v is None:
            v_G[0] = 0.0

    elif truncation == '0D':
        from gpaw.hybrids.wstc import WignerSeitzTruncatedCoulomb
        wstc = WignerSeitzTruncatedCoulomb(qpd.gd.cell_cv, np.ones(3, int))
        v_G = wstc.get_potential(qpd)

    elif truncation in {'1D'}:
        raise feature_removed()

    else:
        raise ValueError('Truncation scheme %s not implemented' % truncation)

    return v_G.astype(complex)


def calculate_2D_truncated_coulomb(qpd, q_v=None, N_c=None):
    """ Simple 2D truncation of Coulomb kernel PRB 73, 205119.
    The non-periodic direction is determined from k-point grid.
    """

    qG_Gv = qpd.get_reciprocal_vectors(add_q=True)
    if qpd.kd.gamma:
        if q_v is not None:
            qG_Gv += q_v
        else:  # only to avoid warning. Later set to zero in factory function
            qG_Gv[0] = [1., 1., 1.]

    # The non-periodic direction is determined from k-point grid
    Nn_c = np.where(N_c == 1)[0]
    Np_c = np.where(N_c != 1)[0]
    if len(Nn_c) != 1:
        # The k-point grid does not fit with boundary conditions
        Nn_c = [2]  # choose reduced cell vectors 0, 1
        Np_c = [0, 1]  # choose reduced cell vector 2
    # Truncation length is half of cell vector in non-periodic direction
    R = qpd.gd.cell_cv[Nn_c[0], Nn_c[0]] / 2.

    qGp_G = ((qG_Gv[:, Np_c[0]])**2 + (qG_Gv[:, Np_c[1]]**2))**0.5
    qGn_G = qG_Gv[:, Nn_c[0]]

    v_G = 4 * np.pi / (qG_Gv**2).sum(axis=1)
    if np.allclose(qGn_G[0], 0) or qpd.kd.gamma:
        """sin(qGn_G * R) = 0 when R = L/2 and q_n = 0.0"""
        v_G *= 1.0 - np.exp(-qGp_G * R) * np.cos(qGn_G * R)
    else:
        """Normal component of q is not zero"""
        a_G = qGn_G / qGp_G * np.sin(qGn_G * R) - np.cos(qGn_G * R)
        v_G *= 1. + np.exp(-qGp_G * R) * a_G

    return v_G.astype(complex)


def get_integrated_kernel(qpd, N_c, truncation=None, N=100, reduced=False):
    from scipy.special import j1, k0, j0, k1

    B_cv = 2 * np.pi * qpd.gd.icell_cv
    Nf_c = np.array([N, N, N])
    if reduced:
        # Only integrate periodic directions if truncation is used
        Nf_c[np.where(N_c == 1)[0]] = 1
    q_qc = monkhorst_pack(Nf_c) / N_c
    q_qc += qpd.q_c
    q_qv = np.dot(q_qc, B_cv)

    if truncation is None:
        V_q = 4 * np.pi / np.sum(q_qv**2, axis=1)
    elif truncation == '2D':
        # The non-periodic direction is determined from k-point grid
        Nn_c = np.where(N_c == 1)[0]
        Np_c = np.where(N_c != 1)[0]
        if len(Nn_c) != 1:
            # The k-point grid does not fit with boundary conditions
            Nn_c = [2]  # choose reduced cell vectors 0, 1
            Np_c = [0, 1]  # choose reduced cell vector 2
        # Truncation length is half of cell vector in non-periodic direction
        R = qpd.gd.cell_cv[Nn_c[0], Nn_c[0]] / 2.

        qp_q = ((q_qv[:, Np_c[0]])**2 + (q_qv[:, Np_c[1]]**2))**0.5
        qn_q = q_qv[:, Nn_c[0]]

        V_q = 4 * np.pi / (q_qv**2).sum(axis=1)
        a_q = qn_q / qp_q * np.sin(qn_q * R) - np.cos(qn_q * R)
        V_q *= 1. + np.exp(-qp_q * R) * a_q
    elif truncation == '1D':
        # The non-periodic direction is determined from k-point grid
        Nn_c = np.where(N_c == 1)[0]
        Np_c = np.where(N_c != 1)[0]

        if len(Nn_c) != 2:
            # The k-point grid does not fit with boundary conditions
            Nn_c = [0, 1]    # Choose reduced cell vectors 0, 1
            Np_c = [2]       # Choose reduced cell vector 2
        # The radius is determined from area of non-periodic part of cell
        Acell_cv = qpd.gd.cell_cv[Nn_c, :][:, Nn_c]
        R = abs(np.linalg.det(Acell_cv) / np.pi)**0.5

        qnR_q = (q_qv[:, Nn_c[0]]**2 + q_qv[:, Nn_c[1]]**2)**0.5 * R
        qpR_q = abs(q_qv[:, Np_c[0]]) * R
        V_q = 4 * np.pi / (q_qv**2).sum(axis=1)
        V_q *= (1.0 + qnR_q * j1(qnR_q) * k0(qpR_q)
                - qpR_q * j0(qnR_q) * k1(qpR_q))
    elif truncation == '0D':
        R = (3 * qpd.gd.volume / (4 * np.pi))**(1. / 3.)
        q2_q = (q_qv**2).sum(axis=1)
        V_q = 4 * np.pi / q2_q
        V_q *= 1.0 - np.cos(q2_q**0.5 * R)

    return np.sum(V_q) / len(V_q), np.sum(V_q**0.5) / len(V_q)


def feature_removed():
    return RuntimeError(
        '0D and 1D truncation have been removed due to not being tested.  '
        'If you need them, please find them in '
        'ec9e49e25613bb99cd69eec9d2613e38b9f6e6e1 '
        'and make sure to add tests in order to have them re-added.')



import os

import numpy as np
import pytest
from ase import Atoms

from gpaw import PW, FermiDirac
from gpaw.calculator import GPAW
from gpaw.new.calculation import DFTCalculation


@pytest.mark.stress
def test_pw_augment_grids(in_tmp_dir):
    ecut = 200
    kpoints = [1, 1, 4]
    atoms = Atoms('HLi', cell=[6, 6, 3.4], pbc=True,
                  positions=[[3, 3, 0], [3, 3, 1.6]])

    def calculate(aug):
        if os.environ.get('GPAW_NEW'):
            dft = DFTCalculation.from_parameters(
                atoms,
                dict(mode=PW(ecut),
                     txt=f'gpaw.aug{aug}.txt',
                     parallel={'augment_grids': aug},
                     kpts={'size': kpoints},
                     occupations=FermiDirac(width=0.1)))
            dft.converge(steps=4)
            e = dft.energies()
            f = dft.forces()
            s = dft.stress()
        else:
            atoms.calc = GPAW(mode=PW(ecut),
                              txt=f'gpaw.aug{aug}.txt',
                              parallel={'augment_grids': aug},
                              kpts={'size': kpoints},
                              occupations=FermiDirac(width=0.1),
                              convergence={'maximum iterations': 4})
            e = atoms.get_potential_energy()
            f = atoms.get_forces()
            s = atoms.get_stress()
        return e, f, s

    e1, f1, s1 = calculate(False)
    e2, f2, s2 = calculate(True)

    eerr = abs(e2 - e1)
    ferr = np.abs(f2 - f1).max()
    serr = np.abs(s2 - s1).max()
    if atoms.calc.wfs.world.rank == 0:
        print('errs', eerr, ferr, serr)
    assert eerr < 5e-12, f'bad energy: err={eerr}'
    assert ferr < 5e-12, f'bad forces: err={ferr}'
    assert serr < 5e-12, f'bad stress: err={serr}'



"""Check TDDFT ionizations with Yukawa potential."""
import pytest
from ase import Atoms
from ase.units import Hartree

import _gpaw
from gpaw import GPAW
from gpaw.cluster import Cluster
from gpaw.eigensolvers import RMMDIIS
from gpaw.lrtddft import LrTDDFT
from gpaw.mpi import world
from gpaw.occupations import FermiDirac
from gpaw.test import equal


@pytest.mark.hybrids
def test_rsf_yukawa_lrtddft_short(in_tmp_dir):
    libxc_version = getattr(_gpaw, 'libxc_version', '2.x.y')
    if int(libxc_version.split('.')[0]) < 3:
        from unittest import SkipTest
        raise SkipTest

    o_plus = Cluster(Atoms('Be', positions=[[0, 0, 0]]))
    o_plus.set_initial_magnetic_moments([1.0])
    o_plus.minimal_box(2.5, h=0.35)

    def get_paw(**kwargs):
        """Return calculator object."""
        c = {'energy': 0.05, 'eigenstates': 0.05, 'density': 0.05}
        return GPAW(convergence=c, eigensolver=RMMDIIS(),
                    nbands=3,
                    xc='PBE',
                    parallel={'domain': world.size}, h=0.35,
                    occupations=FermiDirac(width=0.0, fixmagmom=True),
                    **kwargs)

    calc_plus = get_paw(txt='Be_plus_PBE.log', charge=1)
    o_plus.calc = calc_plus
    o_plus.get_potential_energy()

    calc_plus = calc_plus.new(xc='LCY-PBE:omega=0.83:unocc=True',
                              experimental={'niter_fixdensity': 2},
                              txt='Be_plus_LCY_PBE_083.log')
    o_plus.calc = calc_plus
    o_plus.get_potential_energy()

    lr = LrTDDFT(calc_plus, txt='LCY_TDDFT_Be.log',
                 restrict={'istart': 0, 'jend': 1})
    equal(lr.xc.omega, 0.83)
    lr.write('LCY_TDDFT_Be.ex.gz')
    e_ion = 9.3
    ip_i = 13.36
    # reading is problematic with EXX on more than one core
    if world.rank == 0:
        lr2 = LrTDDFT.read('LCY_TDDFT_Be.ex.gz')
        lr2.diagonalize()
        equal(lr2.xc.omega, 0.83)
        ion_i = lr2[0].get_energy() * Hartree + e_ion
        equal(ion_i, ip_i, 0.3)



import scipy.sparse.linalg as sla


class ScipyLinearSolver:
    """Wrapper class for the linear solvers in ``scipy.sparse.linalg``.

    Solve the linear system of equations

            A * x = b

    where A is a linear operator and b is the known rhs. The linear operator
    provided as argument in the ``solve`` method must have a ``shape``
    attribute (a tuble (M,N) where M and N give the size of the corresponding
    matrix) and a ``dtype`` attribute giving datatype of the matrix entries.

    """

    # Supported solvers
    solvers = {'cg': sla.cg,  # symmetric positive definite
               'minres': sla.minres,  # symmetric indefinite
               'gmres': sla.gmres,  # non-symmetric
               'bicg': sla.bicg,  # non-symmetric
               'cgs': sla.cgs,  # similar to bicg
               'bicgstab': sla.bicgstab,
               'qmr': sla.qmr,
               # 'lgmres': sla.lgmres, # scipy v. 0.8.0
               }

    def __init__(self, method='gmres', preconditioner=None, tolerance=1e-5,
                 max_iter=100):
        """Initialize the linear solver.

        Parameters
        ----------
        method: str
            One of the supported linear solvers in scipy.
        preconditioner: LinearOperator
            Instance of 1) class ``LinearOperator`` from the
            ``scipy.sparse.linalg`` package, or 2) user defined class with
            required attributes.

        """

        if method not in ScipyLinearSolver.solvers:
            raise RuntimeError("Unsupported solver %s" % method)

        self.solver = ScipyLinearSolver.solvers[method]
        self.pc = preconditioner

        self.tolerance = tolerance
        self.max_iter = max_iter

        # Iteration counter
        self.i = None

    def solve(self, A, x_nG, b_nG):
        """Solve linear system Ax = b.

        Parameters
        ----------
        A: LinearOperator
            Instance of 1) class ``LinearOperator`` from the
            ``scipy.sparse.linalg`` package, or 2) user defined class with
            required attributes.
        x_nG: ndarray
            Vector to be solved for (must contain initial starting vector on
            input).
        b_nG: ndarray
            Vector with the right-hand side of the linear equation.

        """

        assert x_nG.shape == b_nG.shape
        shape = x_nG.shape

        # Initialize iteration counter
        self.i = 0

        # Reshape arrays for scipy
        x_0 = x_nG.ravel()
        b = b_nG.ravel()

        x, info = self.solver(A, b, x0=x_0, maxiter=self.max_iter, M=self.pc,
                              tol=self.tolerance, callback=self.iteration)

        x_nG[:] = x.reshape(shape)

        return self.i, info

    def iteration(self, x_i):
        """Passed as callback function to the scipy-routine."""

        # Increment iterator counter
        self.i += 1



from math import pi
from typing import Tuple
import numpy as np
from ase.units import Ha

from gpaw.typing import Array1D
from gpaw.setup import Setup


def ekin(dataset: Setup) -> Tuple[Array1D, Array1D, float]:
    """Calculate PAW kinetic energy contribution as a function of G."""
    ds = dataset
    rgd = dataset.rgd
    de_j = ds.data.e_kin_jj.diagonal()
    phit_j = ds.pseudo_partial_waves_j
    e0 = -ds.Kc
    e_k: Array1D = 0.0  # type: ignore

    for f, l, de, phit in zip(ds.f_j, ds.l_j, de_j, phit_j):
        if f == 0.0:
            continue
        phit_r = np.array([phit(r) for r in rgd.r_g])
        G_k, phit_k = rgd.fft(phit_r * rgd.r_g**(l + 1), l)
        e_k += f * 0.5 * phit_k**2 * G_k**4 / (2 * pi)**3
        e0 -= f * de

    return G_k, e_k, e0


def dekindecut(G: Array1D, de: Array1D, ecut: float) -> float:
    """Linear interpolation."""
    dG = G[1]
    G0 = (2 * ecut)**0.5
    g = int(G0 / dG)
    dedG = np.polyval(np.polyfit(G[g:g + 2], de[g:g + 2], 1), G0)
    dedecut = dedG / G0
    return dedecut


if __name__ == '__main__':
    import argparse
    import matplotlib.pyplot as plt
    from gpaw.setup import create_setup

    parser = argparse.ArgumentParser(
        description='Calculate approximation to the energy variation with '
        'plane-wave cutoff energy.  The approximation is to use the kinetic '
        'energy from a PAW atom, which can be calculated efficiently on '
        'a radial grid.')
    parser.add_argument('-d', '--derivative', type=float, metavar='ECUT',
                        help='Calculate derivative of energy correction with '
                        'respect to plane-wave cutoff energy.')
    parser.add_argument('name', help='Name of PAW dataset.')
    args = parser.parse_args()

    ds = create_setup(args.name)

    G, de, e0 = ekin(ds)
    dG = G[1]

    if args.derivative:
        dedecut = -dekindecut(G, de, args.derivative / Ha)
        print('de/decut({}, {} eV) = {:.6f}'
              .format(args.name, args.derivative, dedecut))
    else:
        de = (np.add.accumulate(de) - 0.5 * de[0] - 0.5 * de) * dG

        ecut = 0.5 * G**2 * Ha
        y = (de[-1] - de) * Ha
        plt.plot(ecut, y)
        plt.xlim(300, 1000)
        plt.ylim(0, y[ecut > 300][0])
        plt.show()



import pytest

from contextlib import contextmanager
from pathlib import Path

from ase.build import molecule
from gpaw import GPAW
from gpaw.lrtddft2 import LrTDDFT2


@contextmanager
def no_error():
    yield


pytestmark = pytest.mark.usefixtures('module_tmp_path')


@pytest.fixture(scope='module')
def ground_state():
    atoms = molecule('H2O')
    atoms.center(vacuum=4)

    calc = GPAW(h=0.4, mode='lcao', basis='sz(dzp)', txt='gs.out',
                nbands=6, xc='LDA')
    atoms.calc = calc
    atoms.get_potential_energy()
    gpw_fpath = Path('gs.gpw').resolve()
    calc.write(gpw_fpath, mode='all')
    return gpw_fpath


@pytest.mark.lrtddft
@pytest.mark.parametrize('max_energy_diff, expectation', [
    (15, pytest.raises(RuntimeError, match=r'.* HOMO \(n=3\) .*')),
    (10, no_error())])
def test_max_energy_diff(ground_state, max_energy_diff, expectation,
                         in_tmp_dir):
    calc = GPAW(ground_state)
    with expectation:
        LrTDDFT2('lr2', calc, fxc='LDA', max_energy_diff=max_energy_diff)


@pytest.mark.lrtddft
@pytest.mark.parametrize('min_occ', [None, 0])
@pytest.mark.parametrize('min_unocc', [None, 0])
@pytest.mark.parametrize('max_occ', [None, 5])
@pytest.mark.parametrize('max_unocc', [None, 5])
def test_indices_with_max_energy_diff(ground_state, min_occ, min_unocc,
                                      max_occ, max_unocc, in_tmp_dir):
    calc = GPAW(ground_state)

    if (min_occ is None or min_unocc is None
        or max_occ is None or max_unocc is None):
        expectation = pytest.raises(RuntimeError)
    else:
        expectation = no_error()

    with expectation:
        LrTDDFT2('lr2', calc, fxc='LDA', max_energy_diff=15,
                 min_occ=min_occ, min_unocc=min_unocc,
                 max_occ=max_occ, max_unocc=max_unocc)


@pytest.mark.lrtddft
@pytest.mark.parametrize('min_occ', [None, 0])
@pytest.mark.parametrize('min_unocc', [None, 0])
@pytest.mark.parametrize('max_occ', [None, 5])
@pytest.mark.parametrize('max_unocc', [None, 5])
def test_indices(ground_state, min_occ, min_unocc,
                 max_occ, max_unocc, in_tmp_dir):
    calc = GPAW(ground_state)
    LrTDDFT2('lr2', calc, fxc='LDA',
             min_occ=min_occ, min_unocc=min_unocc,
             max_occ=max_occ, max_unocc=max_unocc)



from ase.cli.run import Runner, str2dict, CLICommand as ASECLICommand

from gpaw import GPAW
from gpaw.mixer import Mixer, MixerSum
from gpaw.occupations import (FermiDirac, MethfesselPaxton,
                              MarzariVanderbilt)
from gpaw import PW


class GPAWRunner(Runner):
    def __init__(self):
        Runner.__init__(self)
        self.calculator_name = 'gpaw'

    def parse(self, args):
        args.calculator = 'gpaw'
        return Runner.parse(self, args)

    def set_calculator(self, atoms, name):
        parameter_namespace = {
            'PW': PW,
            'FermiDirac': FermiDirac,
            'MethfesselPaxton': MethfesselPaxton,
            'MarzariVanderbilt': MarzariVanderbilt,
            'Mixer': Mixer,
            'MixerSum': MixerSum}
        parameters = str2dict(self.args.parameters, parameter_namespace)
        txt = parameters.pop('txt', self.get_filename(name, 'txt'))
        atoms.calc = GPAW(txt=txt, **parameters)

    def calculate(self, atoms, name):
        data = Runner.calculate(self, atoms, name)
        if self.args.write:
            atoms.calc.write(self.args.write)
        if self.args.write_all:
            atoms.calc.write(self.args.write_all, 'all')
        return data


class CLICommand:
    """Run calculation with GPAW.

    Four types of calculations can be done:

    * single point
    * atomic relaxations
    * unit cell + atomic relaxations
    * equation-of-state

    Examples of the four types of calculations:

        gpaw run -p xc=PBE h2o.xyz
        gpaw run -p xc=PBE h2o.xyz -f 0.01
        gpaw run -p "xc=PBE,kpts={density:4}" cu.traj -s 0.01
        gpaw run -p "xc=PBE,kpts={density:4}" cu.traj -E 5,2.0
    """

    @staticmethod
    def add_arguments(parser):
        ASECLICommand.add_more_arguments(parser)
        parser.add_argument('--dry-run', type=int, default=0,
                            metavar='NCPUS',
                            help='Dry run on NCPUS cpus.')
        parser.add_argument('-w', '--write', help='Write gpw-file.')
        parser.add_argument('-W', '--write-all',
                            help='Write gpw-file with wave functions.')

    @staticmethod
    def run(args):
        runner = GPAWRunner()
        runner.parse(args)
        runner.run()



import pytest
from gpaw.test.fuzz import main


@pytest.mark.serial
@pytest.mark.ci
@pytest.mark.parametrize('mode', ['pw', 'lcao', 'fd'])
@pytest.mark.parametrize('pbc', [0, 1])
def test_fuzz(in_tmp_dir, mode, pbc):
    error = main(f'h2 -m {mode} -c new,old -v 4.0 -p {pbc}')
    assert error == 0, mode



from math import pi
import numpy as np
from scipy.spatial.transform import Rotation
"""
The point groups are written as classes that inherit the class
Pointgroup and contain the information about
- the character table in irreducible form
- representations that correspond to translations in x, y and z
- whether the point group contains complex representations

About complex point groups:
The classes of complex groups (such as Th) are tagged
as that they contain the attribute `complex`. For those groups,
only the real character table is given which is used as such
in the reduction formula. For the representations of these
character tables, the regular orthogonality rules between the
rows and columns do not apply due to the lost information about
the imaginary parts.

"""


def rotation(vec):
    rot = Rotation.from_rotvec(vec)
    try:
        return rot.as_matrix()  # new in scipy-1.4.0
    except AttributeError:
        return rot.as_dcm()  # will be removed in scipy-1.6.0


class Pointgroup:
    def unit(self, data):
        return np.eye(3)

    def rotate_mainaxis(self, angle, data=None):
        return rotation([0, 0, angle / 180 * pi])

    def rotate(self, angle, data=None, axis='z'):
        """
        Parameters
        ----------
        data : array_like
            3D data (a wave function)
        angle : float
            Rotational angle in degrees
        axis : string, optional
            The Rotational axis ('x', 'y' or 'z')
        """
        vec = [0, 0, 0]
        vec['xyz'.index(axis)] = angle / 180 * pi
        return rotation(vec).dot(data)

    def C2prime(self, data, angle):
        """180 degree rotation around the secondary axis"""
        newdata = self.rotate(angle=angle, data=data, axis='z')
        newdata2 = self.rotate(angle=180., data=newdata, axis='x')
        return self.rotate(angle=-angle, data=newdata2,
                           axis='z')  # rotate back

    def S(self, angle, data):
        """
        Rotate and mirror
        """
        rotated = self.rotate(angle=angle, data=data, axis='z')
        return self.mirror_xy(rotated)

    def invert(self, data):
        return self.mirror_xy(self.mirror_yz(self.mirror_xz(data)))

    def mirror_xy(self, data):
        return np.array([[1, 0, 0], [0, 1, 0], [0, 0, -1.0]]).dot(data)

    def mirror_xz(self, data):
        return np.array([[1, 0, 0], [0, -1.0, 0], [0, 0, 1]]).dot(data)

    def mirror_yz(self, data):
        return np.array([[-1.0, 0, 0], [0, 1, 0], [0, 0, 1]]).dot(data)


class D5(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('1C5', self.rotate_mainaxis(angle=72.)),
                           ('4C5', self.rotate_mainaxis(angle=-72.)),
                           ('2C5', self.rotate_mainaxis(angle=2 * 72.)),
                           ('3C5', self.rotate_mainaxis(angle=-2 * 72.)),
                           ('C2_0', self.C2(angle=0 * 72.)),
                           ('C2_1', self.C2(angle=1 * 72.)),
                           ('C2_2', self.C2(angle=2 * 72.)),
                           ('C2_3', self.C2(angle=3 * 72.)),
                           ('C2_4', self.C2(angle=4 * 72.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A1', 'A2', 'E1', 'E2']
        self.character_table = [
            [1., 1., 1., 1.], [1., 1., 1., -1.],
            [2., 2 * np.cos(2 * np.pi / 5.), 2 * np.cos(4 * np.pi / 5.), 0.],
            [2., 2 * np.cos(4 * np.pi / 5.), 2 * np.cos(2 * np.pi / 5.), 0.]
        ]
        self.nof_operations = [1, 2, 2, 5]
        self.Tx_i = 2  # row index for representation of translation vector x
        self.Ty_i = 2
        self.Tz_i = 1

    def __str__(self):
        return 'D5'

    def C2(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.C2prime(data, angle)

        if data is None:
            return do_it
        return do_it(data)


class D5h(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('1C5', self.rotate_mainaxis(angle=72.)),
                           ('4C5', self.rotate_mainaxis(angle=-72.)),
                           ('2C5', self.rotate_mainaxis(angle=2 * 72.)),
                           ('3C5', self.rotate_mainaxis(angle=-2 * 72.)),
                           ('C2_0', self.C2(angle=0 * 72.)),
                           ('C2_1', self.C2(angle=1 * 72.)),
                           ('C2_2', self.C2(angle=2 * 72.)),
                           ('C2_3', self.C2(angle=3 * 72.)),
                           ('C2_4', self.C2(angle=4 * 72.)),
                           ('sigma_h', self.sigma_h),
                           ('S5_1', self.S5(angle=72.)),
                           ('S5_2', self.S5(angle=-72.)),
                           ('S53_1', self.S53(angle=72.)),
                           ('S53_2', self.S53(angle=-72.)),
                           ('sigma_v0', self.sigma_v(angle=0.)),
                           ('sigma_v1', self.sigma_v(angle=72.)),
                           ('sigma_v4', self.sigma_v(angle=-72.)),
                           ('sigma_v2', self.sigma_v(angle=2 * 72.)),
                           ('sigma_v3', self.sigma_v(angle=-2 * 72.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = [
            'A\'1', 'A\'2', 'E\'1', 'E\'2', 'A\'\'1', 'A\'\'2', 'E\'\'1',
            'E\'\'2'
        ]
        self.character_table = [
            [1., 1., 1., 1., 1., 1., 1., 1.],
            [1., 1., 1., -1., 1., 1., 1., -1.],
            [2., 2 * np.cos(2 * np.pi / 5.),
             2 * np.cos(4 * np.pi / 5.), 0., 2.,
             2 * np.cos(2 * np.pi / 5.),
             2 * np.cos(4 * np.pi / 5.), 0.],
            [2.,
             2 * np.cos(4 * np.pi / 5.),
             2 * np.cos(2 * np.pi / 5.),
             0.,
             2.,
             2 * np.cos(4 * np.pi / 5.),
             2 * np.cos(2 * np.pi / 5.),
             0.],
            [1., 1., 1., 1., -1., -1., -1., -1.],
            [1., 1., 1., -1., -1., -1., -1., 1.],
            [2., 2 * np.cos(2 * np.pi / 5.),
             2 * np.cos(4 * np.pi / 5.), 0., -2.,
             -2 * np.cos(2 * np.pi / 5.),
             -2 * np.cos(4 * np.pi / 5.), 0.],
            [2.,
             2 * np.cos(4 * np.pi / 5.),
             2 * np.cos(2 * np.pi / 5.),
             0.,
             -2.,
             -2 * np.cos(4 * np.pi / 5.),
             -2 * np.cos(2 * np.pi / 5.),
             0.]]
        self.nof_operations = [1, 2, 2, 5, 1, 2, 2, 5]
        self.Tx_i = 2
        self.Ty_i = 2
        self.Tz_i = 5

    def __str__(self):
        return 'D5h'

    def C2(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.C2prime(data, angle)

        if data is None:
            return do_it
        return do_it(data)

    def sigma_h(self, data=None):
        return self.mirror_xy(data)

    def S5(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.S(angle=angle, data=data)

        if data is None:
            return do_it
        return do_it(data)

    def S53(self, angle, data=None):
        angle = angle

        def do_it(data):
            data = self.S(angle=angle, data=data)
            data = self.S(angle=angle, data=data)
            data = self.S(angle=angle, data=data)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma_v(self, angle, data=None):
        # first rotate so that the plane is xz plane, flip, and rotate back
        angle = angle

        def do_it(data):
            first = self.rotate(data=data, angle=angle)
            second = self.mirror_xz(first)
            third = self.rotate(data=second, angle=-angle)
            return third

        if data is None:
            return do_it
        return do_it(data)


class Ih(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        # some geometric math:
        a = 1.
        r_cap = np.sin(2 * np.pi / 5) * a
        r_side = np.sqrt(3) * (3 + np.sqrt(5)) / 12. * a
        r_edge = np.cos(np.pi / 5) * a
        r_captoside = 1 / np.sqrt(3) * a
        r_captoedge = 0.5 * a
        r_height = np.sqrt(3) / 2. * a
        self.angle_captocap = np.arccos(((r_cap**2 + r_cap**2) - a**2) /
                                        (2 * r_cap * r_cap)) * 180. / np.pi
        self.angle_captoside = np.arccos(
            ((r_side**2 + r_cap**2) - r_captoside**2) /
            (2 * r_side * r_cap)) * 180. / np.pi
        self.angle_captoedge = np.arccos(
            ((r_edge**2 + r_cap**2) - r_captoedge**2) /
            (2 * r_edge * r_cap)) * 180. / np.pi
        self.angle_captoedge2 = np.arccos(
            ((r_edge**2 + r_cap**2) - r_height**2) /
            (2 * r_edge * r_cap)) * 180. / np.pi
        self.angle_sidetoside = 180. - (2 * self.angle_captoside +
                                        self.angle_captocap)

        self.operations = [
            ('E', self.unit), ('01C5', self.rotate_mainaxis(angle=72.)),
            ('02C5', self.rotate_mainaxis(angle=-72.)),
            ('03C5', self.rotate_othermainaxis(numberofcap=0, angle=72.)),
            ('04C5', self.rotate_othermainaxis(numberofcap=0, angle=-72.)),
            ('05C5', self.rotate_othermainaxis(numberofcap=1, angle=72.)),
            ('06C5', self.rotate_othermainaxis(numberofcap=1, angle=-72.)),
            ('07C5', self.rotate_othermainaxis(numberofcap=2, angle=72.)),
            ('08C5', self.rotate_othermainaxis(numberofcap=2, angle=-72.)),
            ('09C5', self.rotate_othermainaxis(numberofcap=3, angle=72.)),
            ('10C5', self.rotate_othermainaxis(numberofcap=3, angle=-72.)),
            ('11C5', self.rotate_othermainaxis(numberofcap=4, angle=72.)),
            ('12C5', self.rotate_othermainaxis(numberofcap=4, angle=-72.)),
            ('01C52', self.rotate_mainaxis(angle=2 * 72.)),
            ('02C52', self.rotate_mainaxis(angle=-2 * 72.)),
            ('03C52', self.rotate_othermainaxis(numberofcap=0, angle=2 * 72.)),
            ('04C52', self.rotate_othermainaxis(numberofcap=0,
                                                angle=-2 * 72.)),
            ('05C52', self.rotate_othermainaxis(numberofcap=1, angle=2 * 72.)),
            ('06C52', self.rotate_othermainaxis(numberofcap=1,
                                                angle=-2 * 72.)),
            ('07C52', self.rotate_othermainaxis(numberofcap=2, angle=2 * 72.)),
            ('08C52', self.rotate_othermainaxis(numberofcap=2,
                                                angle=-2 * 72.)),
            ('09C52', self.rotate_othermainaxis(numberofcap=3, angle=2 * 72.)),
            ('10C52', self.rotate_othermainaxis(numberofcap=3,
                                                angle=-2 * 72.)),
            ('11C52', self.rotate_othermainaxis(numberofcap=4, angle=2 * 72.)),
            ('12C52', self.rotate_othermainaxis(numberofcap=4,
                                                angle=-2 * 72.)),
            ('01C3', self.rotate_C3(numberofside=0, angle=120.)),
            ('02C3', self.rotate_C3(numberofside=0, angle=-120.)),
            ('03C3', self.rotate_C3(numberofside=1, angle=120.)),
            ('04C3', self.rotate_C3(numberofside=1, angle=-120.)),
            ('05C3', self.rotate_C3(numberofside=2, angle=120.)),
            ('06C3', self.rotate_C3(numberofside=2, angle=-120.)),
            ('07C3', self.rotate_C3(numberofside=3, angle=120.)),
            ('08C3', self.rotate_C3(numberofside=3, angle=-120.)),
            ('09C3', self.rotate_C3(numberofside=4, angle=120.)),
            ('10C3', self.rotate_C3(numberofside=4, angle=-120.)),
            ('11C3', self.rotate_C3(numberofside=5, angle=120.)),
            ('12C3', self.rotate_C3(numberofside=5, angle=-120.)),
            ('13C3', self.rotate_C3(numberofside=6, angle=120.)),
            ('14C3', self.rotate_C3(numberofside=6, angle=-120.)),
            ('15C3', self.rotate_C3(numberofside=7, angle=120.)),
            ('16C3', self.rotate_C3(numberofside=7, angle=-120.)),
            ('17C3', self.rotate_C3(numberofside=8, angle=120.)),
            ('18C3', self.rotate_C3(numberofside=8, angle=-120.)),
            ('19C3', self.rotate_C3(numberofside=9, angle=120.)),
            ('20C3', self.rotate_C3(numberofside=9, angle=-120.)),
            ('01C2', self.C2(numberofedge=0)), ('02C2',
                                                self.C2(numberofedge=1)),
            ('03C2', self.C2(numberofedge=2)), ('04C2',
                                                self.C2(numberofedge=3)),
            ('05C2', self.C2(numberofedge=4)), ('06C2',
                                                self.C2(numberofedge=5)),
            ('07C2', self.C2(numberofedge=6)), ('08C2',
                                                self.C2(numberofedge=7)),
            ('09C2', self.C2(numberofedge=8)),
            ('10C2', self.C2(numberofedge=9)),
            ('11C2', self.C2(numberofedge=10)),
            ('12C2', self.C2(numberofedge=11)),
            ('13C2', self.C2(numberofedge=12)),
            ('14C2', self.C2(numberofedge=13)),
            ('15C2', self.C2(numberofedge=14)), ('i', self.inversion),
            ('01S10', self.S10(angle=36.)), ('02S10', self.S10(angle=-36.)),
            ('03S10', self.S10_othercap(numberofcap=0, angle=36.)),
            ('04S10', self.S10_othercap(numberofcap=0, angle=-36.)),
            ('05S10', self.S10_othercap(numberofcap=1, angle=36.)),
            ('06S10', self.S10_othercap(numberofcap=1, angle=-36.)),
            ('07S10', self.S10_othercap(numberofcap=2, angle=36.)),
            ('08S10', self.S10_othercap(numberofcap=2, angle=-36.)),
            ('09S10', self.S10_othercap(numberofcap=3, angle=36.)),
            ('10S10', self.S10_othercap(numberofcap=3, angle=-36.)),
            ('11S10', self.S10_othercap(numberofcap=4, angle=36.)),
            ('12S10', self.S10_othercap(numberofcap=4, angle=-36.)),
            ('01S10_3', self.S10_3(angle=36.)),
            ('02S10_3', self.S10_3(angle=-36.)),
            ('03S10_3', self.S10_3_othercap(numberofcap=0, angle=36.)),
            ('04S10_3', self.S10_3_othercap(numberofcap=0, angle=-36.)),
            ('05S10_3', self.S10_3_othercap(numberofcap=1, angle=36.)),
            ('06S10_3', self.S10_3_othercap(numberofcap=1, angle=-36.)),
            ('07S10_3', self.S10_3_othercap(numberofcap=2, angle=36.)),
            ('08S10_3', self.S10_3_othercap(numberofcap=2, angle=-36.)),
            ('09S10_3', self.S10_3_othercap(numberofcap=3, angle=36.)),
            ('10S10_3', self.S10_3_othercap(numberofcap=3, angle=-36.)),
            ('11S10_3', self.S10_3_othercap(numberofcap=4, angle=36.)),
            ('12S10_3', self.S10_3_othercap(numberofcap=4, angle=-36.)),
            ('01S6', self.S6(numberofside=0, angle=60.)),
            ('02S6', self.S6(numberofside=0, angle=-60.)),
            ('03S6', self.S6(numberofside=1, angle=60.)),
            ('04S6', self.S6(numberofside=1, angle=-60.)),
            ('05S6', self.S6(numberofside=2, angle=60.)),
            ('06S6', self.S6(numberofside=2, angle=-60.)),
            ('07S6', self.S6(numberofside=3, angle=60.)),
            ('08S6', self.S6(numberofside=3, angle=-60.)),
            ('09S6', self.S6(numberofside=4, angle=60.)),
            ('10S6', self.S6(numberofside=4, angle=-60.)),
            ('11S6', self.S6(numberofside=5, angle=60.)),
            ('12S6', self.S6(numberofside=5, angle=-60.)),
            ('13S6', self.S6(numberofside=6, angle=60.)),
            ('14S6', self.S6(numberofside=6, angle=-60.)),
            ('15S6', self.S6(numberofside=7, angle=60.)),
            ('16S6', self.S6(numberofside=7, angle=-60.)),
            ('17S6', self.S6(numberofside=8, angle=60.)),
            ('18S6', self.S6(numberofside=8, angle=-60.)),
            ('19S6', self.S6(numberofside=9, angle=60.)),
            ('20S6', self.S6(numberofside=9, angle=-60.)),
            ('01sigma', self.sigma(numberofedge=0)),
            ('02sigma', self.sigma(numberofedge=1)),
            ('03sigma', self.sigma(numberofedge=2)),
            ('04sigma', self.sigma(numberofedge=3)),
            ('05sigma', self.sigma(numberofedge=4)),
            ('06sigma', self.sigma(numberofedge=5)),
            ('07sigma', self.sigma(numberofedge=6)),
            ('08sigma', self.sigma(numberofedge=7)),
            ('09sigma', self.sigma(numberofedge=8)),
            ('10sigma', self.sigma(numberofedge=9)),
            ('11sigma', self.sigma(numberofedge=10)),
            ('12sigma', self.sigma(numberofedge=11)),
            ('13sigma', self.sigma(numberofedge=12)),
            ('14sigma', self.sigma(numberofedge=13)),
            ('15sigma', self.sigma(numberofedge=14))
        ]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = [
            'Ag', 'T1g', 'T2g', 'Gg', 'Hg', 'Au', 'T1u', 'T2u', 'Gu', 'Hu'
        ]
        self.character_table = [
            [1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
            [
                3., -2 * np.cos(4 * np.pi / 5.), -2 * np.cos(2 * np.pi / 5.),
                0., -1., 3., -2 * np.cos(2 * np.pi / 5.),
                -2 * np.cos(4 * np.pi / 5.), 0., -1.
            ],
            [
                3., -2 * np.cos(2 * np.pi / 5.), -2 * np.cos(4 * np.pi / 5.),
                0., -1., 3., -2 * np.cos(4 * np.pi / 5.),
                -2 * np.cos(2 * np.pi / 5.), 0., -1.
            ], [4., -1., -1., 1., 0., 4., -1., -1., 1., 0.],
            [5., 0., 0., -1., 1., 5., 0., 0., -1., 1.],
            [1., 1., 1., 1., 1., -1., -1., -1., -1., -1.],
            [
                3., -2 * np.cos(4 * np.pi / 5.), -2 * np.cos(2 * np.pi / 5.),
                0., -1., -3., 2 * np.cos(2 * np.pi / 5.),
                2 * np.cos(4 * np.pi / 5.), 0., 1.
            ],
            [
                3., -2 * np.cos(2 * np.pi / 5.), -2 * np.cos(4 * np.pi / 5.),
                0., -1., -3., 2 * np.cos(4 * np.pi / 5.),
                2 * np.cos(2 * np.pi / 5.), 0., 1.
            ], [4., -1., -1., 1., 0., -4., 1., 1., -1., 0.],
            [5., 0., 0., -1., 1., -5., 0., 0., 1., -1.]
        ]
        self.nof_operations = [1, 12, 12, 20, 15, 1, 12, 12, 20, 15]
        self.Tx_i = 6
        self.Ty_i = 6
        self.Tz_i = 6

    def __str__(self):
        return 'Ih'

    def rotate_othermainaxis(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 72., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 72., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def rotate_C3(self, numberofside, angle, data=None):
        angle = angle
        numberofside = numberofside
        n_is_even = (numberofside % 2 == 0)
        # first, bring a face center to z-axis:
        angle1 = int((numberofside + 1e-4) / 2.) * 72.
        if n_is_even:
            angle2 = -(self.angle_captoside + self.angle_sidetoside)
        else:
            angle2 = -self.angle_captoside

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, numberofedge, angle=180., data=None):
        angle = angle
        numberofedge = numberofedge
        n_is_g1 = (numberofedge % 3 == 0)
        n_is_g2 = (numberofedge % 3 == 1)
        # first, bring an edge center to z-axis:
        angle1 = int((numberofedge + 1e-4) / 3.) * 72. + 36.
        if n_is_g1:
            angle2 = 0.
            angle3 = -self.angle_captoedge
        elif n_is_g2:
            angle2 = 36.
            angle3 = -self.angle_captoedge2
        else:
            angle2 = 18.
            angle3 = -90.

        def do_it(data):
            data = self.rotate(angle=angle1 + angle2, data=data, axis='z')
            data = self.rotate(angle=angle3, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle3, data=data, axis='y')
            data = self.rotate(angle=-(angle1 + angle2), data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def inversion(self, data=None):
        return self.invert(data)

    def S10(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.S(data=data, angle=angle)

        if data is None:
            return do_it
        return do_it(data)

    def S10_othercap(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 72., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual S-operation:
            data = self.S(angle=angle, data=data)

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 72., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def S10_3(self, angle, data=None):
        angle = angle

        def do_it(data):
            first = self.S(data=data, angle=angle)
            second = self.S(data=first, angle=angle)
            third = self.S(data=second, angle=angle)
            return third

        if data is None:
            return do_it
        return do_it(data)

    def S10_3_othercap(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 72., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual S-operation:
            data = self.S(data=data, angle=angle)
            data = self.S(data=data, angle=angle)
            data = self.S(data=data, angle=angle)

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 72., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def S6(self, numberofside, angle, data=None):
        angle = angle
        numberofside = numberofside
        n_is_even = (numberofside % 2 == 0)
        # first, bring a face center to z-axis:
        angle1 = int((numberofside + 1e-4) / 2.) * 72.
        if n_is_even:
            angle2 = (self.angle_captoside + self.angle_sidetoside)
        else:
            angle2 = self.angle_captoside

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=-angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.S(angle=angle, data=data)

            # rotate back:
            data = self.rotate(angle=angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma(self, numberofedge, data=None):
        numberofedge = numberofedge
        n_is_g1 = (numberofedge % 3 == 0)
        n_is_g2 = (numberofedge % 3 == 1)
        # first, bring an edge center to z-axis:
        angle1 = int((numberofedge + 1e-4) / 3.) * 72. + 36.
        if n_is_g1:
            angle2 = 0.
            angle3 = self.angle_captoedge
            angle4 = 0.
        elif n_is_g2:
            angle2 = 36.
            angle3 = self.angle_captoedge2
            angle4 = 90.
        else:
            angle2 = 18.
            angle3 = 90.
            angle4 = self.angle_captoedge

        def do_it(data):
            data = self.rotate(angle=angle1 + angle2, data=data, axis='z')
            data = self.rotate(angle=-angle3, data=data, axis='y')
            data = self.rotate(angle=angle4, data=data, axis='z')

            # do the actual rotation:
            data = self.mirror_xz(data)

            # rotate back:
            data = self.rotate(angle=-angle4, data=data, axis='z')
            data = self.rotate(angle=angle3, data=data, axis='y')
            data = self.rotate(angle=-(angle1 + angle2), data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)


class Ico(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        # some geometric math:
        a = 1.
        r_cap = np.sin(2 * np.pi / 5) * a
        r_side = np.sqrt(3) * (3 + np.sqrt(5)) / 12. * a
        r_edge = np.cos(np.pi / 5) * a
        r_captoside = 1 / np.sqrt(3) * a
        r_captoedge = 0.5 * a
        r_height = np.sqrt(3) / 2. * a
        self.angle_captocap = np.arccos(((r_cap**2 + r_cap**2) - a**2) /
                                        (2 * r_cap * r_cap)) * 180. / np.pi
        self.angle_captoside = np.arccos(
            ((r_side**2 + r_cap**2) - r_captoside**2) /
            (2 * r_side * r_cap)) * 180. / np.pi
        self.angle_captoedge = np.arccos(
            ((r_edge**2 + r_cap**2) - r_captoedge**2) /
            (2 * r_edge * r_cap)) * 180. / np.pi
        self.angle_captoedge2 = np.arccos(
            ((r_edge**2 + r_cap**2) - r_height**2) /
            (2 * r_edge * r_cap)) * 180. / np.pi
        self.angle_sidetoside = 180. - (2 * self.angle_captoside +
                                        self.angle_captocap)

        self.operations = [
            ('E', self.unit), ('01C5', self.rotate_mainaxis(angle=72.)),
            ('02C5', self.rotate_mainaxis(angle=-72.)),
            ('03C5', self.rotate_othermainaxis(numberofcap=0, angle=72.)),
            ('04C5', self.rotate_othermainaxis(numberofcap=0, angle=-72.)),
            ('05C5', self.rotate_othermainaxis(numberofcap=1, angle=72.)),
            ('06C5', self.rotate_othermainaxis(numberofcap=1, angle=-72.)),
            ('07C5', self.rotate_othermainaxis(numberofcap=2, angle=72.)),
            ('08C5', self.rotate_othermainaxis(numberofcap=2, angle=-72.)),
            ('09C5', self.rotate_othermainaxis(numberofcap=3, angle=72.)),
            ('10C5', self.rotate_othermainaxis(numberofcap=3, angle=-72.)),
            ('11C5', self.rotate_othermainaxis(numberofcap=4, angle=72.)),
            ('12C5', self.rotate_othermainaxis(numberofcap=4, angle=-72.)),
            ('01C52', self.rotate_mainaxis(angle=2 * 72.)),
            ('02C52', self.rotate_mainaxis(angle=-2 * 72.)),
            ('03C52', self.rotate_othermainaxis(numberofcap=0, angle=2 * 72.)),
            ('04C52', self.rotate_othermainaxis(numberofcap=0,
                                                angle=-2 * 72.)),
            ('05C52', self.rotate_othermainaxis(numberofcap=1, angle=2 * 72.)),
            ('06C52', self.rotate_othermainaxis(numberofcap=1,
                                                angle=-2 * 72.)),
            ('07C52', self.rotate_othermainaxis(numberofcap=2, angle=2 * 72.)),
            ('08C52', self.rotate_othermainaxis(numberofcap=2,
                                                angle=-2 * 72.)),
            ('09C52', self.rotate_othermainaxis(numberofcap=3, angle=2 * 72.)),
            ('10C52', self.rotate_othermainaxis(numberofcap=3,
                                                angle=-2 * 72.)),
            ('11C52', self.rotate_othermainaxis(numberofcap=4, angle=2 * 72.)),
            ('12C52', self.rotate_othermainaxis(numberofcap=4,
                                                angle=-2 * 72.)),
            ('01C3', self.rotate_C3(numberofside=0, angle=120.)),
            ('02C3', self.rotate_C3(numberofside=0, angle=-120.)),
            ('03C3', self.rotate_C3(numberofside=1, angle=120.)),
            ('04C3', self.rotate_C3(numberofside=1, angle=-120.)),
            ('05C3', self.rotate_C3(numberofside=2, angle=120.)),
            ('06C3', self.rotate_C3(numberofside=2, angle=-120.)),
            ('07C3', self.rotate_C3(numberofside=3, angle=120.)),
            ('08C3', self.rotate_C3(numberofside=3, angle=-120.)),
            ('09C3', self.rotate_C3(numberofside=4, angle=120.)),
            ('10C3', self.rotate_C3(numberofside=4, angle=-120.)),
            ('11C3', self.rotate_C3(numberofside=5, angle=120.)),
            ('12C3', self.rotate_C3(numberofside=5, angle=-120.)),
            ('13C3', self.rotate_C3(numberofside=6, angle=120.)),
            ('14C3', self.rotate_C3(numberofside=6, angle=-120.)),
            ('15C3', self.rotate_C3(numberofside=7, angle=120.)),
            ('16C3', self.rotate_C3(numberofside=7, angle=-120.)),
            ('17C3', self.rotate_C3(numberofside=8, angle=120.)),
            ('18C3', self.rotate_C3(numberofside=8, angle=-120.)),
            ('19C3', self.rotate_C3(numberofside=9, angle=120.)),
            ('20C3', self.rotate_C3(numberofside=9, angle=-120.)),
            ('01C2', self.C2(numberofedge=0)), ('02C2',
                                                self.C2(numberofedge=1)),
            ('03C2', self.C2(numberofedge=2)), ('04C2',
                                                self.C2(numberofedge=3)),
            ('05C2', self.C2(numberofedge=4)), ('06C2',
                                                self.C2(numberofedge=5)),
            ('07C2', self.C2(numberofedge=6)), ('08C2',
                                                self.C2(numberofedge=7)),
            ('09C2', self.C2(numberofedge=8)),
            ('10C2', self.C2(numberofedge=9)),
            ('11C2', self.C2(numberofedge=10)),
            ('12C2', self.C2(numberofedge=11)),
            ('13C2', self.C2(numberofedge=12)),
            ('14C2', self.C2(numberofedge=13)),
            ('15C2', self.C2(numberofedge=14))
        ]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A', 'T1', 'T2', 'G', 'H']
        self.character_table = [
            [1., 1., 1., 1., 1.],
            [3., -2 * np.cos(4 * np.pi / 5.),
             -2 * np.cos(2 * np.pi / 5.), 0., -1.],
            [3., -2 * np.cos(2 * np.pi / 5.),
             -2 * np.cos(4 * np.pi / 5.), 0., -1.],
            [4., -1., -1., 1., 0.],
            [5., 0., 0., -1., 1.]]
        self.nof_operations = [1, 12, 12, 20, 15]
        self.Tx_i = 1
        self.Ty_i = 1
        self.Tz_i = 1

    def __str__(self):
        return 'Ico'

    def rotate_othermainaxis(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 72., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 72., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def rotate_C3(self, numberofside, angle, data=None):
        angle = angle
        numberofside = numberofside
        n_is_even = (numberofside % 2 == 0)
        # first, bring a face center to z-axis:
        angle1 = int((numberofside + 1e-4) / 2.) * 72.
        if n_is_even:
            angle2 = -(self.angle_captoside + self.angle_sidetoside)
        else:
            angle2 = -self.angle_captoside

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, numberofedge, angle=180., data=None):
        angle = angle
        numberofedge = numberofedge
        n_is_g1 = (numberofedge % 3 == 0)
        n_is_g2 = (numberofedge % 3 == 1)
        # first, bring an edge center to z-axis:
        angle1 = int((numberofedge + 1e-4) / 3.) * 72. + 36.
        if n_is_g1:
            angle2 = 0.
            angle3 = -self.angle_captoedge
        elif n_is_g2:
            angle2 = 36.
            angle3 = -self.angle_captoedge2
        else:
            angle2 = 18.
            angle3 = -90.

        def do_it(data):
            data = self.rotate(angle=angle1 + angle2, data=data, axis='z')
            data = self.rotate(angle=angle3, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle3, data=data, axis='y')
            data = self.rotate(angle=-(angle1 + angle2), data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)


class Td(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        # some geometric math:
        # a = 1.
        # r_cap = np.sqrt(3/8.) * a
        # r_face = r_cap/3.
        # r_edge = np.sqrt(r_cap * r_face)
        # r_captoside = 1/np.sqrt(3) * a
        # r_captoedge = 0.5*a
        # r_height = np.sqrt(3)/2. * a
        self.angle_captoedge = np.arctan(np.sqrt(2)) * 180. / np.pi
        self.angle_captocap = 2 * self.angle_captoedge
        self.angle_captoface = np.arctan(2. * np.sqrt(2)) * 180. / np.pi

        self.operations = [
            ('E', self.unit), ('1C3', self.rotate_mainaxis(angle=120.)),
            ('2C3', self.rotate_mainaxis(angle=-120.)),
            ('3C3', self.rotate_othermainaxis(numberofcap=0, angle=120.)),
            ('4C3', self.rotate_othermainaxis(numberofcap=0, angle=-120.)),
            ('5C3', self.rotate_othermainaxis(numberofcap=1, angle=120.)),
            ('6C3', self.rotate_othermainaxis(numberofcap=1, angle=-120.)),
            ('7C3', self.rotate_othermainaxis(numberofcap=2, angle=120.)),
            ('8C3', self.rotate_othermainaxis(numberofcap=2, angle=-120.)),
            ('1C2', self.C2(numberofedge=0)), ('2C2', self.C2(numberofedge=1)),
            ('3C2', self.C2(numberofedge=2)),
            ('1S4', self.S4(numberofside=0, angle=90.)),
            ('2S4', self.S4(numberofside=0, angle=-90.)),
            ('3S4', self.S4(numberofside=1, angle=90.)),
            ('4S4', self.S4(numberofside=1, angle=-90.)),
            ('5S4', self.S4(numberofside=2, angle=90.)),
            ('6S4', self.S4(numberofside=2, angle=-90.)),
            ('1sigma', self.sigma(numberofedge=0)),
            ('2sigma', self.sigma(numberofedge=1)),
            ('3sigma', self.sigma(numberofedge=2)),
            ('4sigma', self.sigma(numberofedge=3)),
            ('5sigma', self.sigma(numberofedge=4)),
            ('6sigma', self.sigma(numberofedge=5))
        ]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A1', 'A2', 'E', 'T1', 'T2']
        self.character_table = [[1., 1., 1., 1., 1.], [1., 1., 1., -1., -1.],
                                [2., -1., 2., 0., 0.], [3., 0., -1., 1., -1.],
                                [3., 0., -1., -1., 1.]]
        self.nof_operations = [1, 8, 3, 6, 6]
        self.Tx_i = 4
        self.Ty_i = 4
        self.Tz_i = 4

    def __str__(self):
        return 'Td'

    def rotate_othermainaxis(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 120., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 120., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, numberofedge, angle=180., data=None):
        angle = angle
        numberofedge = numberofedge
        # first, bring an edge center to z-axis:
        angle1 = numberofedge * 120.
        angle2 = self.angle_captoedge

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def S4(self, numberofside, angle, data=None):
        angle = angle
        numberofside = numberofside
        # first, bring an edge center to z-axis:
        angle1 = numberofside * 120.
        angle2 = self.angle_captoedge

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.S(angle=angle, data=data)

            # rotate back:
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma(self, numberofedge, data=None):
        numberofedge = numberofedge
        # first, bring an edge center to z-axis:
        angle1 = (numberofedge % 3) * 120.
        angle2 = self.angle_captoedge
        angle3 = (numberofedge % 2) * 90.

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')
            data = self.rotate(angle=angle3, data=data, axis='z')

            # do the actual reflection:
            data = self.mirror_xz(data)

            # rotate back:
            data = self.rotate(angle=-angle3, data=data, axis='z')
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)


class Th(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        # some geometric math:
        # a = 1.
        # r_cap = np.sqrt(3/8.) * a
        # r_face = r_cap/3.
        # r_edge = np.sqrt(r_cap * r_face)
        # r_captoside = 1/np.sqrt(3) * a
        # r_captoedge = 0.5*a
        # r_height = np.sqrt(3)/2. * a
        self.angle_captoedge = np.arctan(np.sqrt(2)) * 180. / np.pi
        self.angle_captocap = 2 * self.angle_captoedge
        self.angle_captoface = np.arctan(2. * np.sqrt(2)) * 180. / np.pi

        self.operations = [
            ('E', self.unit), ('1C3', self.rotate_mainaxis(angle=120.)),
            ('2C3', self.rotate_othermainaxis(numberofcap=0, angle=120.)),
            ('3C3', self.rotate_othermainaxis(numberofcap=1, angle=120.)),
            ('4C3', self.rotate_othermainaxis(numberofcap=2, angle=120.)),
            ('1C3_2', self.rotate_mainaxis(angle=-120.)),
            ('2C3_2', self.rotate_othermainaxis(numberofcap=0, angle=-120.)),
            ('3C3_2', self.rotate_othermainaxis(numberofcap=1, angle=-120.)),
            ('4C3_2', self.rotate_othermainaxis(numberofcap=2, angle=-120.)),
            ('1C2', self.C2(numberofedge=0)), ('2C2', self.C2(numberofedge=1)),
            ('3C2', self.C2(numberofedge=2)), ('i', self.inversion),
            ('1S6_5', self.S6(angle=-60.)),
            ('2S6_5', self.S6_othercap(numberofcap=0, angle=-60.)),
            ('3S6_5', self.S6_othercap(numberofcap=1, angle=-60.)),
            ('4S6_5', self.S6_othercap(numberofcap=2, angle=-60.)),
            ('5S6', self.S6(angle=60.)),
            ('6S6', self.S6_othercap(numberofcap=0, angle=60.)),
            ('7S6', self.S6_othercap(numberofcap=1, angle=60.)),
            ('8S6', self.S6_othercap(numberofcap=2, angle=60.)),
            ('1sigma', self.sigma(numberofedge=0)),
            ('2sigma', self.sigma(numberofedge=1)),
            ('3sigma', self.sigma(numberofedge=2))
        ]

        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['Ag', 'Eg', 'Tg', 'Au', 'Eu', 'Tu']
        self.character_table = [
            [1., 1., 1., 1., 1., 1.],
            [2., 2. * np.cos(2 * np.pi / 3.), 2., 2.,
             2. * np.cos(2 * np.pi / 3.), 2.],
            [3., 0., -1., 3., 0., -1.],
            [1., 1., 1., -1., -1., -1.],
            [2., 2. * np.cos(2 * np.pi / 3.), 2., -2.,
             -2. * np.cos(2 * np.pi / 3.), -2.],
            [3., 0., -1., -3., 0., 1.]]
        self.nof_operations = [1, 8, 3, 1, 8, 3]
        self.Tx_i = 5
        self.Ty_i = 5
        self.Tz_i = 5
        self.complex = True

    def __str__(self):
        return 'Th'

    def rotate_othermainaxis(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            # first, bring another cap to z-axis:
            data = self.rotate(angle=numberofcap * 120., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 120., data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, numberofedge, angle=180., data=None):
        angle = angle
        numberofedge = numberofedge
        # first, bring an edge center to z-axis:
        angle1 = numberofedge * 120.
        angle2 = self.angle_captoedge

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')

            # do the actual rotation:
            data = self.rotate(angle=angle, data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def inversion(self, data=None):
        return self.invert(data)

    def S6(self, angle, data=None):
        angle = angle

        def do_it(data):
            data = self.S(angle=angle, data=data)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def S6_othercap(self, numberofcap, angle, data=None):
        angle = angle
        numberofcap = numberofcap

        def do_it(data):
            data = self.rotate(angle=numberofcap * 120., data=data, axis='z')
            data = self.rotate(angle=self.angle_captocap, data=data, axis='y')

            # do the actual rotation:
            data = self.S(angle=angle, data=data)

            # rotate back:
            data = self.rotate(angle=-self.angle_captocap, data=data, axis='y')
            data = self.rotate(angle=-numberofcap * 120, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma(self, numberofedge, data=None):
        numberofedge = numberofedge
        # first, bring an edge center to z-axis:
        angle1 = (numberofedge % 3) * 120.
        angle2 = self.angle_captoedge
        angle3 = 45.

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='z')
            data = self.rotate(angle=angle2, data=data, axis='y')
            data = self.rotate(angle=angle3, data=data, axis='z')

            # do the actual reflection:
            data = self.mirror_xz(data)

            # rotate back:
            data = self.rotate(angle=-angle3, data=data, axis='z')
            data = self.rotate(angle=-angle2, data=data, axis='y')
            data = self.rotate(angle=-angle1, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)


class C2v(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('C2', self.rotate_mainaxis(angle=180.)),
                           ('sigma_v0', self.sigma_v(angle=0.)),
                           ('sigma_v1', self.sigma_v(angle=90.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A1', 'A2', 'B1', 'B2']
        self.character_table = [[1., 1., 1., 1.], [1., 1., -1., -1.],
                                [1., -1., 1., -1.], [1., -1., -1., 1.]]
        self.nof_operations = [1, 1, 1, 1]
        self.Tx_i = 2
        self.Ty_i = 3
        self.Tz_i = 0

    def __str__(self):
        return 'C2v'

    def C2(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.C2prime(data, angle)

        if data is None:
            return do_it
        return do_it(data)

    def sigma_v(self, angle, data=None):
        # first rotate so that the plane is xz plane, flip, and rotate back
        angle = angle

        def do_it(data):
            first = self.rotate(data=data, angle=angle)
            second = self.mirror_xz(first)
            third = self.rotate(data=second, angle=-angle)
            return third

        if data is None:
            return do_it
        return do_it(data)


class C3v(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('1C3', self.rotate_mainaxis(angle=120.)),
                           ('2C3', self.rotate_mainaxis(angle=-120.)),
                           ('sigma_v0', self.sigma_v(angle=0.)),
                           ('sigma_v1', self.sigma_v(angle=120.)),
                           ('sigma_v2', self.sigma_v(angle=-120.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A1', 'A2', 'E']
        self.character_table = [[1., 1., 1.], [1., 1., -1.], [2., -1., 0.]]
        self.nof_operations = [1, 2, 3]
        self.Tx_i = 2
        self.Ty_i = 2
        self.Tz_i = 0

    def __str__(self):
        return 'C3v'

    def sigma_v(self, angle, data=None):
        # first rotate so that the plane is xz plane, flip, and rotate back
        angle = angle

        def do_it(data):
            first = self.rotate(data=data, angle=angle)
            second = self.mirror_xz(first)
            third = self.rotate(data=second, angle=-angle)
            return third

        if data is None:
            return do_it
        return do_it(data)


class D3h(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('1C3', self.rotate_mainaxis(angle=120.)),
                           ('2C3', self.rotate_mainaxis(angle=-120.)),
                           ('C2_0', self.C2(angle=0 * 120.)),
                           ('C2_1', self.C2(angle=1 * 120.)),
                           ('C2_2', self.C2(angle=2 * 120.)),
                           ('sigma_h', self.sigma_h),
                           ('S3_1', self.S3(angle=120.)),
                           ('S3_2', self.S3(angle=-120.)),
                           ('sigma_v0', self.sigma_v(angle=0.)),
                           ('sigma_v1', self.sigma_v(angle=120.)),
                           ('sigma_v2', self.sigma_v(angle=-120.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = [
            'A\'1', 'A\'2', 'E\'1', 'A\'\'1', 'A\'\'2', 'E\'\'1'
        ]
        self.character_table = [[1., 1., 1., 1., 1., 1.],
                                [1., 1., -1., 1., 1., -1.],
                                [2., -1., 0., 2., -1., 0.],
                                [1., 1., 1., -1., -1., -1.],
                                [1., 1., -1., -1., -1., 1.],
                                [2., -1., 0., -2., 1., 0.]]
        self.nof_operations = [1, 2, 3, 1, 2, 3]
        self.Tx_i = 2
        self.Ty_i = 2
        self.Tz_i = 4

    def __str__(self):
        return 'D3h'

    def C2(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.C2prime(data, angle)

        if data is None:
            return do_it
        return do_it(data)

    def sigma_h(self, data=None):
        return self.mirror_xy(data)

    def S3(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.S(data=data, angle=angle)

        if data is None:
            return do_it
        return do_it(data)

    def S53(self, angle, data=None):
        angle = angle

        def do_it(data):
            first = self.S(data=data, angle=angle)
            second = self.S(data=first, angle=angle)
            third = self.S(data=second, angle=angle)
            return third

        if data is None:
            return do_it
        return do_it(data)

    def sigma_v(self, angle, data=None):
        # first rotate so that the plane is xz plane, flip, and rotate back
        angle = angle

        def do_it(data):
            first = self.rotate(data=data, angle=angle)
            second = self.mirror_xz(first)
            third = self.rotate(data=second, angle=-angle)
            return third

        if data is None:
            return do_it
        return do_it(data)


class D2d(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('S4_1', self.S4(angle=90.)),
                           ('S4_2', self.S4(angle=-90.)),
                           ('C2_0', self.C2(angle=180.)),
                           ("C2'_0", self.C2p(angle=45.)),
                           ("C2'_1", self.C2p(angle=-45.)),
                           ('sigma_d1', self.sigma_d(angle=0.)),
                           ('sigma_d2', self.sigma_d(angle=90.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A\'1', 'A\'2', 'B\'1', 'B\'2', 'E']
        self.character_table = [[1., 1., 1., 1., 1.], [1., 1., 1., -1., -1.],
                                [1., -1., 1., 1., -1.], [1., -1., 1., -1., 1.],
                                [2., 0., -2., 0., 0.]]
        self.nof_operations = [1, 2, 1, 2, 2]
        self.Tx_i = 4
        self.Ty_i = 4
        self.Tz_i = 3

    def __str__(self):
        return 'D2d'

    def S4(self, angle, data=None):
        angle = angle

        # first, bring an edge center to z-axis:

        def do_it(data):
            # do the actual rotation:
            data = self.S(angle=angle, data=data)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, angle, data=None):
        angle = angle

        def do_it(data):
            return self.rotate(angle=angle, data=data, axis='z')

        if data is None:
            return do_it
        return do_it(data)

    def C2p(self, angle=90., data=None):
        angle = angle

        def do_it(data):
            data = self.rotate(angle=angle, data=data, axis='z')
            data = self.rotate(angle=90., data=data, axis='x')

            # do the actual rotation:
            data = self.rotate(angle=180., data=data, axis='z')

            # rotate back:
            data = self.rotate(angle=-90., data=data, axis='x')
            data = self.rotate(angle=-angle, data=data, axis='z')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma_d(self, angle, data=None):
        # first rotate so that the plane is xz plane, flip, and rotate back
        angle = angle

        def do_it(data):
            first = self.rotate(data=data, angle=angle)
            second = self.mirror_xz(first)
            third = self.rotate(data=second, angle=-angle)
            return third

        if data is None:
            return do_it
        return do_it(data)


class C2(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit),
                           ('C2', self.rotate_mainaxis(angle=180.))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = ['A', 'B']
        self.character_table = [[
            1.,
            1.,
        ], [1., -1.]]

        self.nof_operations = [1, 1]
        self.Tx_i = 1
        self.Ty_i = 1
        self.Tz_i = 0

    def __str__(self):
        return 'C2'


class Oh(Pointgroup):
    # main axis should be the z-axis!
    def __init__(self):
        self.operations = [('E', self.unit), ('1C3', self.C3(corner=0)),
                           ('2C3', self.C3(corner=1)),
                           ('3C3', self.C3(corner=2)),
                           ('4C3', self.C3(corner=3)),
                           ('5C3', self.C3(corner=4)),
                           ('6C3', self.C3(corner=5)),
                           ('7C3', self.C3(corner=6)),
                           ('8C3', self.C3(corner=7)),
                           ('1C2', self.C2(edge=0)), ('2C2', self.C2(edge=1)),
                           ('3C2', self.C2(edge=2)), ('4C2', self.C2(edge=3)),
                           ('5C2', self.C2(edge=4)), ('6C2', self.C2(edge=5)),
                           ('1C4', self.C4(face=0)), ('2C4', self.C4(face=1)),
                           ('3C4', self.C4(face=2)), ('4C4', self.C4(face=3)),
                           ('5C4', self.C4(face=4)), ('6C4', self.C4(face=5)),
                           ('1faceC2', self.faceC2(face=0)),
                           ('2faceC2', self.faceC2(face=1)),
                           ('3faceC2', self.faceC2(face=2)),
                           ('i', self.inversion), ('1S4', self.S4(face=0)),
                           ('2S4', self.S4(face=1)), ('3S4', self.S4(face=2)),
                           ('4S4', self.S4(face=3)), ('5S4', self.S4(face=4)),
                           ('6S4', self.S4(face=5)),
                           ('1S6', self.S6(corner=0)),
                           ('2S6', self.S6(corner=1)),
                           ('3S6', self.S6(corner=2)),
                           ('4S6', self.S6(corner=3)),
                           ('5S6', self.S6(corner=4)),
                           ('6S6', self.S6(corner=5)),
                           ('7S6', self.S6(corner=6)),
                           ('8S6', self.S6(corner=7)),
                           ('1sigma_h1', self.sigma_h(face=0)),
                           ('2sigma_h1', self.sigma_h(face=1)),
                           ('3sigma_h1', self.sigma_h(face=2)),
                           ('1sigma_d1', self.sigma_d(face=0)),
                           ('2sigma_d1', self.sigma_d(face=1)),
                           ('3sigma_d1', self.sigma_d(face=2)),
                           ('4sigma_d1', self.sigma_d(face=3)),
                           ('5sigma_d1', self.sigma_d(face=4)),
                           ('6sigma_d1', self.sigma_d(face=5))]
        self.operation_names = [pair[0] for pair in self.operations]
        self.symmetries = [
            'A1g', 'A2g', 'Eg', 'T1g', 'T2g', 'A1u', 'A2u', 'Eu', 'T1u', 'T2u'
        ]
        self.character_table = [[1., 1., 1., 1., 1., 1., 1., 1., 1., 1.],
                                [1., 1., -1., -1., 1., 1., -1., 1., 1., -1.],
                                [2., -1., 0., 0., 2., 2., 0., -1., 2., 0.],
                                [3., 0., -1., 1., -1., 3., 1., 0., -1., -1.],
                                [3., 0., 1., -1., -1., 3., -1., 0., -1., 1.],
                                [1., 1., 1., 1., 1., -1., -1., -1., -1., -1.],
                                [1., 1., -1., -1., 1., -1., 1., -1., -1., 1.],
                                [2., -1., 0., 0., 2., -2., 0., 1., -2., 0.],
                                [3., 0., -1., 1., -1., -3., -1., 0., 1., 1.],
                                [3., 0., 1., -1., -1., -3., 1., 0., 1., -1.]]
        self.nof_operations = [1, 8, 6, 6, 3, 1, 6, 8, 3, 6]
        self.Tx_i = 8
        self.Ty_i = 8
        self.Tz_i = 8

    def __str__(self):
        return 'Oh'

    def C3(self, corner, data=None):
        corner = corner
        t1, t2 = divmod(corner, 4)
        t1 = (-1)**t1
        angle1 = (t2 * 90.) + 45.  # y
        angle2 = t1 * (np.arctan(1 / np.sqrt(2.)) * 180. / np.pi)  # x

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='y')
            data = self.rotate(angle=angle2, data=data, axis='x')

            data = self.rotate(angle=120., data=data, axis='z')

            data = self.rotate(angle=-angle2, data=data, axis='x')
            data = self.rotate(angle=-angle1, data=data, axis='y')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C2(self, edge, data=None):
        edge = edge
        t1, t2 = divmod(edge, 3)
        angle1 = t1 * 90.
        set2 = [[45., 'x'], [45., 'y'], [-45., 'x']][t2]

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='y')
            data = self.rotate(angle=set2[0], data=data, axis=set2[1])

            data = self.rotate(angle=180., data=data, axis='z')

            data = self.rotate(angle=-set2[0], data=data, axis=set2[1])
            data = self.rotate(angle=-angle1, data=data, axis='y')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def C4(self, face, data=None):
        face = face
        t1, t2 = divmod(face, 3)
        axis = ['x', 'y', 'z'][t2]
        angle = (-1)**t1 * 90.

        def do_it(data):
            data = self.rotate(angle=90., data=data, axis=axis)

            data = self.rotate(angle=angle, data=data, axis='z')

            data = self.rotate(angle=-90., data=data, axis=axis)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def faceC2(self, face, data=None):
        face = face
        t1, t2 = divmod(face, 3)
        axis = ['x', 'y', 'z'][t2]
        angle = 180.

        def do_it(data):
            data = self.rotate(angle=90., data=data, axis=axis)

            data = self.rotate(angle=angle, data=data, axis='y')

            data = self.rotate(angle=-90., data=data, axis=axis)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def inversion(self, data=None):
        return self.invert(data)

    def S4(self, face, data=None):
        face = face
        t1, t2 = divmod(face, 3)
        axis = ['x', 'y', 'z'][t2]
        angle = (-1)**t1 * 90.

        def do_it(data):
            data = self.rotate(angle=90., data=data, axis=axis)

            data = self.S(angle=angle, data=data)

            data = self.rotate(angle=-90., data=data, axis=axis)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def S6(self, corner, data=None):
        corner = corner
        t1, t2 = divmod(corner, 4)
        t1 = (-1)**t1
        angle1 = (t2 * 90.) + 45.  # y
        angle2 = t1 * (np.arctan(1 / np.sqrt(2.)) * 180. / np.pi)  # x

        def do_it(data):
            data = self.rotate(angle=angle1, data=data, axis='y')
            data = self.rotate(angle=angle2, data=data, axis='x')

            data = self.S(angle=60., data=data)

            data = self.rotate(angle=-angle2, data=data, axis='x')
            data = self.rotate(angle=-angle1, data=data, axis='y')
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma_h(self, face, data=None):
        face = face
        t1, t2 = divmod(face, 3)
        axis = ['x', 'y', 'z'][t2]

        def do_it(data):
            data = self.rotate(angle=90., data=data, axis=axis)

            data = self.mirror_xy(data)

            data = self.rotate(angle=-90., data=data, axis=axis)
            return data

        if data is None:
            return do_it
        return do_it(data)

    def sigma_d(self, face, data=None):
        face = face
        t1, t2 = divmod(face, 3)
        axis = ['x', 'y', 'z'][t2]
        angle2 = (-1)**t1 * 45.

        def do_it(data):
            data = self.rotate(angle=90., data=data, axis=axis)
            data = self.rotate(angle=angle2, data=data, axis='z')

            data = self.mirror_xz(data)

            data = self.rotate(angle=-angle2, data=data, axis='z')
            data = self.rotate(angle=-90., data=data, axis=axis)
            return data

        if data is None:
            return do_it
        return do_it(data)
